import logging

from furl import furl

from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from mayan.apps.converter.classes import ConverterBase
from mayan.apps.converter.exceptions import AppImageError
from mayan.apps.converter.models import LayerTransformation
from mayan.apps.converter.settings import setting_image_generation_timeout
from mayan.apps.converter.transformations import BaseTransformation
from mayan.apps.file_caching.models import CachePartitionFile
from mayan.apps.lock_manager.backends.base import LockingBackend

from ..literals import IMAGE_ERROR_VERSION_PAGE_TRANSFORMATION_ERROR

logger = logging.getLogger(name=__name__)


class DocumentVersionPageBusinessLogicMixin:
    @cached_property
    def cache_partition(self):
        partition, created = self.document_version.cache.partitions.get_or_create(
            name=self.uuid
        )
        return partition

    def generate_image(
        self, _acquire_lock=True, maximum_layer_order=None,
        transformation_instance_list=None, user=None
    ):
        combined_transformation_list = self.get_combined_transformation_list(
            maximum_layer_order=maximum_layer_order,
            transformation_instance_list=transformation_instance_list,
            user=user
        )
        combined_cache_filename = self.get_combined_cache_filename(
            _combined_transformation_list=combined_transformation_list
        )

        logger.debug(
            'transformations cache filename: %s', combined_cache_filename
        )

        content_object_lock_name = self.content_object.get_lock_name(user=user)
        try:
            content_object_lock = LockingBackend.get_backend().acquire_lock(
                name=content_object_lock_name,
                timeout=setting_image_generation_timeout.value * 2
            )
        except Exception:
            raise
        else:
            lock_name = self.get_lock_name(
                _combined_cache_filename=combined_cache_filename
            )
            try:
                if _acquire_lock:
                    lock = LockingBackend.get_backend().acquire_lock(
                        name=lock_name,
                        timeout=setting_image_generation_timeout.value
                    )
            except Exception:
                raise
            else:
                # Second try block to release the lock even on fatal errors
                # inside the block.
                try:
                    try:
                        self.cache_partition.get_file(
                            filename=combined_cache_filename
                        )
                    except CachePartitionFile.DoesNotExist:
                        logger.debug(
                            'transformations cache file "%s" not found, '
                            'generating new image', combined_cache_filename
                        )
                        image = self.get_image(
                            transformation_instance_list=combined_transformation_list
                        )
                        with self.cache_partition.create_file(filename=combined_cache_filename) as file_object:
                            file_object.write(
                                image.getvalue()
                            )
                    else:
                        logger.debug(
                            'transformations cache file "%s" found, '
                            'returning it to caller', combined_cache_filename
                        )

                    return combined_cache_filename
                finally:
                    if _acquire_lock:
                        lock.release()
            finally:
                content_object_lock.release()

    def get_api_image_url(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None, viewname=None, view_kwargs=None
    ):
        """
        Create an unique URL combining:
        - the page's image URL
        - the interactive argument
        - a hash from the server side and interactive transformations
        The purpose of this unique URL is to allow client side caching
        if document page images.
        """
        if not self.content_object:
            return '#'

        transformation_instance_list = transformation_instance_list or ()

        # Source object transformations first.
        transformation_list = LayerTransformation.objects.get_for_object(
            as_classes=True, maximum_layer_order=maximum_layer_order,
            obj=self.content_object, user=user
        )

        transformation_list.extend(
            self.get_combined_transformation_list(
                maximum_layer_order=maximum_layer_order,
                transformation_instance_list=transformation_instance_list,
                user=user
            )
        )
        try:
            transformations_hash = BaseTransformation.combine(
                transformations=transformation_list
            )
        except Exception as exception:
            raise AppImageError(
                error_name=IMAGE_ERROR_VERSION_PAGE_TRANSFORMATION_ERROR
            ) from exception
        else:
            view_kwargs = view_kwargs or {
                'document_id': self.document_version.document_id,
                'document_version_id': self.document_version_id,
                'document_version_page_id': self.pk
            }

            final_url = furl()
            final_url.path = reverse(
                viewname=viewname or 'rest_api:documentversionpage-image',
                kwargs=view_kwargs
            )
            # Remove leading '?' character.
            final_url.query = BaseTransformation.list_as_query_string(
                transformation_instance_list=transformation_instance_list
            )[1:]
            final_url.args['_hash'] = transformations_hash

            if maximum_layer_order is not None:
                final_url.args['maximum_layer_order'] = maximum_layer_order

            return final_url.tostr()

    def get_combined_cache_filename(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None, _combined_transformation_list=None
    ):
        combined_transformation_list = _combined_transformation_list or self.get_combined_transformation_list(
            maximum_layer_order=maximum_layer_order,
            transformation_instance_list=transformation_instance_list,
            user=user
        )

        content_object_cache_filename = self.content_object.get_combined_cache_filename(
            user=user
        )
        return '{}-{}'.format(
            content_object_cache_filename,
            BaseTransformation.combine(
                transformations=combined_transformation_list
            )
        )

    def get_combined_transformation_list(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None
    ):
        """
        Return a list of transformation containing the server side
        transformations for this object as well as transformations
        created from the arguments as transient interactive transformation.
        """
        result = []

        # Stored transformations first.
        result.extend(
            LayerTransformation.objects.get_for_object(
                as_classes=True, maximum_layer_order=maximum_layer_order,
                obj=self, user=user
            )
        )

        # Interactive transformations second.
        result.extend(
            transformation_instance_list or []
        )

        return result

    def get_image(self, transformation_instance_list=None):
        cache_filename = '{}-base_image'.format(
            self.content_object.get_combined_cache_filename()
        )
        logger.debug('Page cache filename: %s', cache_filename)

        try:
            cache_file = self.cache_partition.get_file(
                filename=cache_filename
            )
        except CachePartitionFile.DoesNotExist:
            logger.debug('Page cache version "%s" not found', cache_filename)

            try:
                content_object_cache_filename = self.content_object.generate_image(
                    _acquire_lock=False
                )
                content_object_cache_file = self.content_object.cache_partition.get_file(
                    filename=content_object_cache_filename
                )

                with content_object_cache_file.open() as file_object:
                    converter = ConverterBase.get_converter_class()(
                        file_object=file_object
                    )
                    converter.seek_page(page_number=0)

                    page_image = converter.get_page()

                    # Since open "wb+" doesn't create versions, create it
                    # explicitly.
                    with self.cache_partition.create_file(filename=cache_filename) as file_object:
                        file_object.write(
                            page_image.getvalue()
                        )

                    # Apply runtime transformations.
                    for transformation in transformation_instance_list or ():
                        converter.transform(transformation=transformation)

                    return converter.get_page()
            except Exception as exception:
                # Cleanup in case of error.
                logger.error(
                    'Error creating document version page cache file '
                    'named "%s"; %s', cache_filename, exception,
                    exc_info=True
                )
                raise
        else:
            logger.debug('Page cache version "%s" found', cache_filename)

            with cache_file.open() as file_object:
                converter = ConverterBase.get_converter_class()(
                    file_object=file_object
                )

                converter.seek_page(page_number=0)

                # This code is also repeated below to allow using a context
                # manager with cache_version.open and close it automatically.
                # Apply runtime transformations.
                for transformation in transformation_instance_list or ():
                    converter.transform(transformation=transformation)

                return converter.get_page()

    def get_label(self):
        return _(
            '%(document_version)s page %(page_number)d of %(total_pages)d'
        ) % {
            'document_version': str(self.document_version),
            'page_number': self.page_number,
            'total_pages': self.get_pages_last_number() or 1
        }
    get_label.short_description = _('Label')

    def get_lock_name(
        self, _combined_cache_filename=None, maximum_layer_order=None,
        transformation_instance_list=None, user=None
    ):
        if _combined_cache_filename:
            combined_cache_filename = _combined_cache_filename
        else:
            combined_cache_filename = self.get_combined_cache_filename(
                maximum_layer_order=maximum_layer_order,
                transformation_instance_list=transformation_instance_list,
                user=user
            )

        return 'document_version_page_generate_image_{}_{}'.format(
            self.pk, combined_cache_filename
        )

    @property
    def is_in_trash(self):
        return self.document_version.is_in_trash

    @property
    def uuid(self):
        """
        Make cache UUID a mix of version ID and page ID to avoid using stale
        images.
        """
        return '{}-{}'.format(self.document_version.uuid, self.pk)



from rest_framework.generics import get_object_or_404

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.serializers.document_serializers import DocumentSerializer
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from .models import Cabinet
from .permissions import (
    permission_cabinet_add_document, permission_cabinet_create,
    permission_cabinet_delete, permission_cabinet_edit,
    permission_cabinet_remove_document, permission_cabinet_view
)
from .serializers import (
    CabinetDocumentAddSerializer, CabinetDocumentRemoveSerializer,
    CabinetSerializer
)


class APIDocumentCabinetListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    Returns a list of all the cabinets to which a document belongs.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    mayan_external_object_permissions = {
        'GET': (permission_cabinet_view,)
    }
    mayan_object_permissions = {
        'GET': (permission_cabinet_view,)
    }
    serializer_class = CabinetSerializer

    def get_source_queryset(self):
        return self.get_external_object().cabinets.all()


class APICabinetListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the cabinets.
    post: Create a new cabinet.
    """
    mayan_object_permissions = {
        'GET': (permission_cabinet_view,)
    }
    mayan_view_permissions = {
        'POST': (permission_cabinet_create,)
    }
    ordering_fields = ('id', 'label')
    serializer_class = CabinetSerializer
    source_queryset = Cabinet.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_mayan_view_permissions(self, request, view):
        if request.method == 'POST':
            serializer = self.get_serializer(data=request.data)
            serializer.is_valid(raise_exception=True)

            if serializer.validated_data['parent']:
                return ()
            else:
                return self.mayan_view_permissions.get(request.method, None)
        else:
            return self.mayan_view_permissions.get(request.method, None)

    def perform_create(self, serializer):
        parent = serializer.validated_data['parent']

        if parent:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission_cabinet_create,
                queryset=self.get_source_queryset(), user=self.request.user
            )
            get_object_or_404(queryset=queryset, pk=parent.pk)

        return super().perform_create(serializer)


class APICabinetView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected cabinet.
    get: Returns the details of the selected cabinet.
    patch: Edit the selected cabinet.
    put: Edit the selected cabinet.
    """
    lookup_url_kwarg = 'cabinet_id'
    mayan_object_permissions = {
        'GET': (permission_cabinet_view,),
        'PUT': (permission_cabinet_edit,),
        'PATCH': (permission_cabinet_edit,),
        'DELETE': (permission_cabinet_delete,)
    }
    serializer_class = CabinetSerializer
    source_queryset = Cabinet.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APICabinetDocumentAddView(generics.ObjectActionAPIView):
    """
    post: Add a document to a cabinet.
    """
    lookup_url_kwarg = 'cabinet_id'
    mayan_object_permissions = {
        'POST': (permission_cabinet_add_document,)
    }
    serializer_class = CabinetDocumentAddSerializer
    source_queryset = Cabinet.objects.all()

    def object_action(self, obj, request, serializer):
        document = serializer.validated_data['document']
        obj.document_add(document=document, user=self.request.user)


class APICabinetDocumentRemoveView(generics.ObjectActionAPIView):
    """
    post: Remove a document from a cabinet.
    """
    lookup_url_kwarg = 'cabinet_id'
    mayan_object_permissions = {
        'POST': (permission_cabinet_remove_document,)
    }
    serializer_class = CabinetDocumentRemoveSerializer
    source_queryset = Cabinet.objects.all()

    def object_action(self, obj, request, serializer):
        document = serializer.validated_data['document']
        obj.document_remove(document=document, user=self.request.user)


class APICabinetDocumentListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of all the documents contained in a particular cabinet.
    """
    external_object_class = Cabinet
    external_object_pk_url_kwarg = 'cabinet_id'
    mayan_external_object_permissions = {
        'GET': (permission_cabinet_view,)
    }
    mayan_object_permissions = {
        'GET': (permission_document_view,)
    }
    serializer_class = DocumentSerializer

    def get_source_queryset(self):
        return Document.valid.filter(
            pk__in=self.get_external_object().documents.only('pk')
        )



from rest_framework import pagination

from .literals import DEFAULT_PAGE_SIZE_QUERY_PARAMETER
from .settings import setting_maximum_page_size, setting_page_size


class MayanPageNumberPagination(pagination.PageNumberPagination):
    max_page_size = setting_maximum_page_size.value
    page_size = setting_page_size.value
    page_size_query_param = DEFAULT_PAGE_SIZE_QUERY_PARAMETER



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('motd', '0004_auto_20160314_0040')
    ]

    operations = [
        migrations.AlterField(
            model_name='message',
            name='end_datetime',
            field=models.DateTimeField(
                help_text='Date and time until when this message is to be '
                'displayed.', null=True, verbose_name='End date time',
                blank=True
            )
        ),
        migrations.AlterField(
            model_name='message',
            name='label',
            field=models.CharField(
                help_text='Short description of this message.',
                max_length=32, verbose_name='Label'
            )
        ),
        migrations.AlterField(
            model_name='message',
            name='message',
            field=models.TextField(
                help_text='The actual message to be displayed.',
                verbose_name='Message'
            )
        ),
        migrations.AlterField(
            model_name='message',
            name='start_datetime',
            field=models.DateTimeField(
                help_text='Date and time after which this message will be '
                'displayed.', null=True, verbose_name='Start date time',
                blank=True
            )
        )
    ]



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0013_auto_20190423_0810')
    ]

    operations = [
        migrations.CreateModel(
            name='WorkflowTransitionField',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'field_type', models.PositiveIntegerField(
                        choices=[(1, 'Character'), (2, 'Number (Integer)')],
                        verbose_name='Type'
                    )
                ),
                (
                    'name', models.CharField(
                        help_text='The name that will be used to identify '
                        'this field in other parts of the workflow system.',
                        max_length=128, verbose_name='Internal name'
                    )
                ),
                (
                    'label', models.CharField(
                        help_text='The field name that will be shown on '
                        'the user interface.', max_length=128,
                        verbose_name='Label'
                    )
                ),
                (
                    'help_text', models.TextField(
                        blank=True, help_text='An optional message that '
                        'will help users better understand the purpose '
                        'of the field and data to provide.',
                        verbose_name='Help text'
                    )
                ),
                (
                    'required', models.BooleanField(
                        default=False, help_text='Whether this fields '
                        'needs to be filled out or not to proceed.',
                        verbose_name='Required'
                    )
                ),
                (
                    'transition', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='fields',
                        to='document_states.WorkflowTransition',
                        verbose_name='Transition'
                    )
                )
            ],
            options={
                'verbose_name': 'Workflow transition trigger event',
                'verbose_name_plural': 'Workflow transitions trigger events'
            }
        ),
        migrations.AddField(
            model_name='workflowinstance',
            name='context',
            field=models.TextField(blank=True, verbose_name='Backend data')
        ),
        migrations.AddField(
            model_name='workflowinstancelogentry',
            name='extra_data',
            field=models.TextField(blank=True, verbose_name='Extra data')
        ),
        migrations.AlterUniqueTogether(
            name='workflowtransitionfield',
            unique_together={
                ('transition', 'name')
            }
        )
    ]



from django.db import models


class DocumentIndexInstanceNodeManager(models.Manager):
    def get_for(self, document):
        return self.filter(documents=document)


class IndexInstanceManager(models.Manager):
    def delete_empty_nodes(self):
        for index in self.all():
            index.delete_empty_nodes()

    def document_add(self, document):
        for index in self.filter(document_types=document.document_type):
            index.document_add(document=document)

    def document_remove(self, document):
        for index_instance in self.filter(index_template_nodes__index_instance_nodes__documents=document):
            index_instance.document_remove(document=document)


class IndexTemplateManager(models.Manager):
    def get_by_natural_key(self, slug):
        return self.get(slug=slug)

    def rebuild(self):
        for index_template in self.all():
            index_template.rebuild()



from django.utils.translation import ugettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_b

queue_parsing = CeleryQueue(
    name='parsing', label=_('Parsing'), worker=worker_b
)

queue_parsing.add_task_type(
    dotted_path='mayan.apps.document_parsing.tasks.task_parse_document_file',
    label=_('Document file parsing')
)



import json
import logging
from pathlib import Path

import sh

from django.utils.translation import ugettext_lazy as _

from mayan.apps.storage.utils import TemporaryDirectory

from ..classes import FileMetadataDriver
from ..literals import DEFAULT_EXIF_PATH
from ..settings import setting_drivers_arguments

logger = logging.getLogger(name=__name__)


class EXIFToolDriver(FileMetadataDriver):
    label = _('EXIF Tool')
    internal_name = 'exiftool'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.read_settings()

        if self.auto_initialize:
            try:
                self.command_exiftool = sh.Command(path=self.exiftool_path)
            except sh.CommandNotFound:
                self.command_exiftool = None
            else:
                self.command_exiftool = self.command_exiftool.bake('-j')

    def _process(self, document_file):
        if self.command_exiftool:
            with TemporaryDirectory() as temporary_folder:
                path_temporary_file = Path(
                    temporary_folder, Path(document_file.filename).name
                )

                try:
                    with path_temporary_file.open(mode='xb') as temporary_fileobject:
                        document_file.save_to_file(file_object=temporary_fileobject)
                        temporary_fileobject.seek(0)
                        try:
                            output = self.command_exiftool(temporary_fileobject.name)
                        except sh.ErrorReturnCode_1 as exception:
                            result = json.loads(s=exception.stdout)[0]
                            if result.get('Error', '') == 'Unknown file type':
                                # Not a fatal error.
                                return result
                        else:
                            return json.loads(s=output)[0]
                except Exception as exception:
                    logger.error(
                        'Error processing document file: %s; %s',
                        document_file, exception, exc_info=True
                    )
                    raise
        else:
            logger.warning(
                'EXIFTool binary not found, not processing document '
                'file: %s', document_file
            )

    def read_settings(self):
        self.exiftool_path = setting_drivers_arguments.value.get(
            'exif_driver', {}
        ).get('exiftool_path', DEFAULT_EXIF_PATH)


EXIFToolDriver.register(
    mimetypes=('*',)
)



from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave

from .classes import GPGBackend
from .events import event_key_created
from .literals import KEY_TYPE_CHOICES
from .managers import KeyManager
from .model_mixins import KeyBusinessLogicMixin


class Key(ExtraDataModelMixin, KeyBusinessLogicMixin, models.Model):
    """
    Fields:
    * key_type - Will show private or public, the only two types of keys in
    a public key infrastructure, the kind used in Mayan.
    """
    key_data = models.TextField(
        help_text=_('ASCII armored version of the key.'),
        verbose_name=_('Key data')
    )
    creation_date = models.DateTimeField(
        editable=False, verbose_name=_('Creation date')
    )
    expiration_date = models.DateTimeField(
        blank=True, editable=False, null=True,
        verbose_name=_('Expiration date')
    )
    fingerprint = models.CharField(
        editable=False, max_length=40, unique=True,
        verbose_name=_('Fingerprint')
    )
    length = models.PositiveIntegerField(
        editable=False, verbose_name=_('Length')
    )
    algorithm = models.PositiveIntegerField(
        editable=False, verbose_name=_('Algorithm')
    )
    user_id = models.TextField(
        editable=False, verbose_name=_('User ID')
    )
    key_type = models.CharField(
        choices=KEY_TYPE_CHOICES, editable=False, max_length=3,
        verbose_name=_('Type')
    )

    objects = KeyManager()

    class Meta:
        verbose_name = _('Key')
        verbose_name_plural = _('Keys')

    def __str__(self):
        return '{} - {}'.format(self.key_id, self.user_id)

    def clean(self):
        """
        Validate the key before saving.
        """
        import_results = GPGBackend.get_instance().import_key(
            key_data=self.key_data
        )

        if not import_results.count:
            raise ValidationError(
                message=_('Invalid key data')
            )

        queryset = Key.objects.filter(
            fingerprint=import_results.fingerprints[0]
        )

        if queryset.exists():
            raise ValidationError(
                message=_('Key already exists.')
            )

    def get_absolute_url(self):
        return reverse(
            viewname='django_gpg:key_detail', kwargs={'key_id': self.pk}
        )

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_key_created,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        self.introspect_key_data()
        super().save(*args, **kwargs)



from ..classes import Template

from .literals import TEST_TEMPLATE


class DocumentTemplateSandboxViewTestMixin:
    def _request_document_template_sandbox_get_view(self):
        return self.get(
            viewname='templating:document_template_sandbox', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_document_template_sandbox_post_view(self):
        return self.post(
            viewname='templating:document_template_sandbox', kwargs={
                'document_id': self._test_document.pk
            }, data={'template_template': TEST_TEMPLATE}
        )


class TemplateTestMixin:
    def _render_test_template(self, template_string, context=None):
        template = Template(template_string=template_string)
        return template.render(context=context)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_c

queue_mailing = CeleryQueue(
    label=_('Mailing'), name='mailing', worker=worker_c
)

queue_mailing.add_task_type(
    dotted_path='mayan.apps.mailer.tasks.task_send_object',
    label=_('Send object')
)



from django.contrib import admin

from .models import Key


@admin.register(Key)
class KeyAdmin(admin.ModelAdmin):
    list_display = (
        'key_id', 'user_id', 'creation_date', 'expiration_date', 'key_type'
    )
    list_filter = ('key_type',)
    readonly_fields = list_display + ('fingerprint', 'length', 'algorithm')
    search_fields = ('key_id', 'user_id',)



from django.utils.translation import ugettext_lazy as _

from rest_framework import serializers as rest_framework_serializers
from rest_framework.fields import (  # NOQA
    BooleanField, CharField, ChoiceField, DateField, DateTimeField,
    DecimalField, DictField, DurationField, EmailField, Field, FileField,
    FilePathField, FloatField, HiddenField, HStoreField, ImageField,
    IntegerField, IPAddressField, JSONField, ListField, ModelField,
    MultipleChoiceField, ReadOnlyField, RegexField, SerializerMethodField,
    SlugField, TimeField, URLField, UUIDField
)
from rest_framework.relations import (  # NOQA
    HyperlinkedIdentityField, HyperlinkedRelatedField, ManyRelatedField,
    PrimaryKeyRelatedField, RelatedField, SlugRelatedField,
    StringRelatedField
)
from rest_framework.reverse import reverse
from rest_framework.serializers import (
    HyperlinkedModelSerializer as RESTFrameworkHyperlinkedModelSerializer,
    ModelSerializer as RESTFrameworkModelSerializer
)

from .classes import BatchRequestCollection
from .serializer_mixins import (
    CreateOnlyFieldSerializerMixin, DynamicFieldListSerializerMixin
)


class Serializer(
    DynamicFieldListSerializerMixin, rest_framework_serializers.Serializer
):
    """Serializer subclass to add Mayan specific mixins."""


class BatchAPIRequestResponseSerializer(Serializer):
    content = CharField(
        label=_('Content'), read_only=True
    )
    data = JSONField(
        label=_('Data'), read_only=True
    )
    headers = DictField(
        label=_('Headers'), read_only=True
    )
    name = CharField(
        label=_('Name'), read_only=True
    )
    status_code = IntegerField(
        label=_('Status code'), read_only=True
    )
    requests = JSONField(
        label=_('Requests'), style={'base_template': 'textarea.html'},
        write_only=True
    )

    def validate(self, data):
        try:
            BatchRequestCollection(
                request_list=data['requests']
            )
        except Exception as exception:
            raise rest_framework_serializers.ValidationError(
                'Error validating requests; {}'.format(exception)
            )

        return data


class BlankSerializer(Serializer):
    """Serializer for the object action API view."""


class EndpointSerializer(Serializer):
    label = CharField(
        label=_('Label'), read_only=True
    )
    url = SerializerMethodField(
        label=_('URL')
    )

    def get_url(self, instance):
        if instance.viewname:
            return reverse(
                kwargs=instance.kwargs, viewname=instance.viewname,
                request=self.context['request'],
                format=self.context['format']
            )


class HyperlinkedModelSerializer(
    CreateOnlyFieldSerializerMixin, DynamicFieldListSerializerMixin,
    RESTFrameworkHyperlinkedModelSerializer
):
    """HyperlinkedModelSerializer subclass to add Mayan specific mixins."""


class ModelSerializer(
    CreateOnlyFieldSerializerMixin, DynamicFieldListSerializerMixin,
    RESTFrameworkModelSerializer
):
    """ModelSerializer subclass to add Mayan specific mixins."""


class ProjectInformationSerializer(Serializer):
    __author__ = CharField(
        label=_('Author'), read_only=True
    )
    __author_email__ = CharField(
        label=_('Author email'), read_only=True
    )
    __build__ = CharField(
        label=_('Build'), read_only=True
    )
    __build_string__ = CharField(
        label=_('Build string'), read_only=True
    )
    __copyright__ = CharField(
        label=_('Copyright'), read_only=True
    )
    __description__ = CharField(
        label=_('Description'), read_only=True
    )
    __django_version__ = CharField(
        label=_('Django version'), read_only=True
    )
    __license__ = CharField(
        label=_('License'), read_only=True
    )
    __title__ = CharField(
        label=_('Title'), read_only=True
    )
    __version__ = CharField(
        label=_('Version'), read_only=True
    )
    __website__ = CharField(
        label=_('Website'), read_only=True
    )



from django.utils.translation import ugettext_lazy as _


class DependencyEnvironment:
    def __init__(self, label, name, help_text=None, mark_missing=False):
        self.label = label
        self.help_text = help_text
        self.name = name
        self.mark_missing = mark_missing

    def __str__(self):
        return str(self.label)


environment_build = DependencyEnvironment(
    help_text=_(
        'Environment used for building distributable packages of the '
        'software. End users can ignore missing dependencies under this '
        'environment.'
    ), label=_('Build'), name='build'
)
environment_development = DependencyEnvironment(
    help_text=_(
        'Environment used for developers to make code changes. End users '
        'can ignore missing dependencies under this environment.'
    ), label=_('Development'), name='development'
)
environment_documentation = DependencyEnvironment(
    help_text=_(
        'Environment used for building the documentation. End users '
        'can ignore missing dependencies under this environment.'
    ), label=_('Documentation'), name='documentation'
)
environment_documentation_override = DependencyEnvironment(
    help_text=_(
        'Environment used to specify direct documentation dependencies to '
        'workaround unpinned or immutable dependency bugs in third party '
        'libraries. End users can ignore missing dependencies under this '
        'environment.'
    ), label=_('Documentation (override)'), name='documentation_override'
)
environment_production = DependencyEnvironment(
    help_text=_(
        'Normal environment for end users. A missing dependency under this '
        'environment will result in issues and errors during normal use.'
    ), label=_('Production'), mark_missing=True, name='production'
)
environment_testing = DependencyEnvironment(
    help_text=_(
        'Environment used running the test suit to verify the functionality '
        'of the code. Dependencies in this environment are not needed for '
        'normal production usage.'
    ), label=_('Testing'), name='testing'
)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link

from .icons import icon_document_version_export
from .permissions import permission_document_version_export

link_document_version_export = Link(
    args='resolved_object.pk', icon=icon_document_version_export,
    permissions=(permission_document_version_export,),
    text=_('Export'), view='document_exports:document_version_export'
)



from django.db import migrations, models

import mayan.apps.common.validators


class Migration(migrations.Migration):
    dependencies = [
        ('converter', '0006_auto_20150708_0120'),
    ]

    operations = [
        migrations.AlterField(
            model_name='transformation',
            name='arguments',
            field=models.TextField(
                help_text='Enter the arguments for the transformation as a '
                'YAML dictionary. ie: {"degrees": 180}', blank=True,
                verbose_name='Arguments',
                validators=[
                    mayan.apps.common.validators.YAMLValidator()
                ]
            ),
            preserve_default=True,
        ),
    ]



import errno
import functools
import os

import yaml

from mayan.apps.common.serialization import yaml_load
from mayan.apps.templating.classes import Template

from .literals import (
    CONFIGURATION_FILENAME, CONFIGURATION_LAST_GOOD_FILENAME
)


class SettingNamespaceSingleton:
    """
    Self hosting bootstrap setting class.
    Allow managing setting in a compatible way before Mayan EDMS starts.
    """
    _setting_kwargs = {}
    _setting_overrides = {}
    _settings = {}

    class SettingNotFound(Exception):
        """Mostly a stand-in or typecast for KeyError for readability."""

    @classmethod
    def load_config_file(cls, filepath):
        try:
            with open(file=filepath) as file_object:
                file_object.seek(0, os.SEEK_END)
                if file_object.tell():
                    file_object.seek(0)
                    try:
                        return yaml_load(stream=file_object)
                    except yaml.YAMLError as exception:
                        exit(
                            'Error loading configuration file: {}; {}'.format(
                                filepath, exception
                            )
                        )
        except IOError as exception:
            if exception.errno == errno.ENOENT:
                # No config file, return empty dictionary
                return {}
            else:
                raise

    @classmethod
    def register_setting(cls, name, klass, kwargs=None):
        cls._settings[name] = klass
        cls._setting_kwargs[name] = kwargs or {}

    def __init__(self, global_symbol_table):
        self.global_symbol_table = global_symbol_table
        self.settings = {}

        for name, klass in self.__class__._settings.items():
            kwargs = self.__class__._setting_kwargs[name].copy()
            kwargs['name'] = name
            setting = klass(**kwargs)
            setting.namespace = self
            self.settings[name] = setting

    @functools.cache
    def get_config_file_content(self):
        filepath = self.get_setting_value(
            name='CONFIGURATION_FILEPATH'
        )

        return self.load_config_file(filepath=filepath) or {}

    def get_setting(self, name):
        return self.settings[name]

    def get_setting_value(self, name):
        """
        Wrapper that calls the individual setting .get_value method.
        Convenience method to allow returning setting values from the
        namespace.
        """
        try:
            setting = self.get_setting(name=name)
        except KeyError:
            raise SettingNamespaceSingleton.SettingNotFound
        else:
            return setting.get_value()

    def get_values(self, only_critical=False):
        """
        Return a dictionary will all the settings and their respective
        resolved values.
        """
        result = {}
        for name, setting in self.settings.items():
            # If only_critical is set to True load only the settings with
            # the critical flag. Otherwise load all.
            if only_critical and setting.critical or not only_critical:
                try:
                    result[name] = setting.get_value()
                except SettingNamespaceSingleton.SettingNotFound:
                    """
                    Not critical, we just avoid adding it to the result
                    dictionary.
                    """
                except Exception as exception:
                    exit(
                        'Unable to load bootstrap setting; {}'.format(
                            exception
                        )
                    )

        return result

    def update_globals(self, global_symbol_table=None, only_critical=False):
        """
        Insert all resolved values into the symbol table of the caller.
        """
        result = self.get_values(only_critical=only_critical)

        if global_symbol_table is None:
            global_symbol_table = self.global_symbol_table

        global_symbol_table.update(result)


class BaseSetting:
    @staticmethod
    def safe_string_value_to_string(value):
        return value.encode('utf-8').decode()

    def __init__(
        self, name, critical=False, default_value=None,
        has_default=False
    ):
        self.critical = critical
        self.default_value = default_value
        self.has_default = has_default
        self.name = name

    def _get_value(self):
        """
        By default will try to get the value from the namespace symbol table,
        then the configuration file, and finally from the environment.
        """
        # Resolution order
        # 1 - Environment
        # 2 - Config
        # 3 - Global
        # 4 - Default
        try:
            return self.load_value_from_environment()
        except SettingNamespaceSingleton.SettingNotFound:
            try:
                return self.load_value_from_config_file()
            except SettingNamespaceSingleton.SettingNotFound:
                try:
                    return self.load_value_from_global_system_table()
                except KeyError:
                    if self.has_default:
                        return self.get_default_value()
                    else:
                        raise SettingNamespaceSingleton.SettingNotFound

    def get_default_value(self):
        return self.default_value

    def get_environment_name(self):
        return 'MAYAN_{}'.format(self.name)

    def get_template_environment_name(self):
        return 'MAYAN_{}'.format(
            self.get_template_name()
        )

    def get_template_name(self):
        return 'SETTING_TEMPLATE_{}'.format(self.name)

    def get_template_string(self):
        try:
            return self.get_template_string_from_environment()
        except KeyError:
            try:
                return self.get_template_string_from_config_file()
            except KeyError:
                return self.get_template_string_from_global_system_table()

    def get_template_string_from_config_file(self):
        self.namespace.get_config_file_content()[
            self.get_template_name()
        ]

    def get_template_string_from_environment(self):
        return os.environ[
            self.get_template_environment_name()
        ]

    def get_template_string_from_global_system_table(self):
        return self.namespace.global_symbol_table[
            self.get_template_name()
        ]

    def get_value(self):
        try:
            return self.namespace._setting_overrides[self.name]
        except KeyError:
            try:
                template_string = self.get_template_string()
            except KeyError:
                return self._get_value()
            else:
                setting_template = Template(template_string=template_string)
                context = {}
                context.update(self.namespace.global_symbol_table)
                context.update(
                    self.namespace.get_config_file_content()
                )
                context.update(os.environ)

                value = setting_template.render(context=context)
                value = BaseSetting.safe_string_value_to_string(value=value)

                return value

    def load_value_from_config_file(self):
        try:
            return self.namespace.get_config_file_content()[self.name]
        except KeyError:
            raise SettingNamespaceSingleton.SettingNotFound

    def load_value_from_environment(self):
        try:
            value = os.environ[
                self.get_environment_name()
            ]
        except KeyError:
            raise SettingNamespaceSingleton.SettingNotFound
        else:
            try:
                return yaml_load(stream=value)
            except yaml.YAMLError as exception:
                raise ValueError(
                    'Error loading setting environment variable "{}", '
                    'value: "{}"; {}'.format(
                        self.name, value, exception
                    )
                )

    def load_value_from_global_system_table(self):
        return self.namespace.global_symbol_table[self.name]

    def set_value(self, value):
        self.namespace._setting_overrides[self.name] = value


class FilesystemBootstrapSetting(BaseSetting):
    def __init__(self, name, critical=False, path_parts=None):
        self.critical = critical
        self.has_default = True
        self.name = name
        self.path_parts = path_parts

    def _get_value(self):
        """
        It is not possible to look for this setting in the config file
        because not even the config file setup has completed.
        This setting only supports being set from the environment.
        """
        try:
            return self.load_value_from_environment()
        except SettingNamespaceSingleton.SettingNotFound:
            if self.has_default:
                return self.get_default_value()
            else:
                raise SettingNamespaceSingleton.SettingNotFound

    def get_default_value(self):
        """
        The default value of this setting class is not static but calculated.
        """
        # Can't use BASE_DIR from django.conf.settings.
        # Use it from the `global_symbol_table` which should be the
        # same.
        try:
            BASE_DIR = self.namespace._setting_overrides['MEDIA_ROOT']
        except KeyError:
            BASE_DIR = self.namespace.global_symbol_table.get(
                'MEDIA_ROOT', self.namespace.global_symbol_table['BASE_DIR']
            )

        return os.path.join(BASE_DIR, *self.path_parts)

    def get_template_string(self):
        try:
            return self.get_template_string_from_environment()
        except KeyError:
            return self.get_template_string_from_global_system_table()


class MediaBootstrapSetting(FilesystemBootstrapSetting):
    def get_default_value(self):
        """
        The default value of this setting class is not static but calculated.
        """
        return os.path.join(
            self.namespace.get_setting_value(name='MEDIA_ROOT'),
            *self.path_parts
        )


def smart_yaml_load(value):
    if isinstance(value, dict):
        return value
    else:
        return yaml_load(
            stream=value or '{}'
        )


# FilesystemBootstrapSetting settings

SettingNamespaceSingleton.register_setting(
    klass=MediaBootstrapSetting, kwargs={
        'critical': True, 'path_parts': (CONFIGURATION_FILENAME,)
    }, name='CONFIGURATION_FILEPATH'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'critical': True, 'has_default': True, 'default_value': False
    }, name='CONFIGURATION_FILE_IGNORE'
)

# MediaBootstrapSetting settings

SettingNamespaceSingleton.register_setting(
    klass=MediaBootstrapSetting, kwargs={
        'critical': True, 'path_parts': (CONFIGURATION_LAST_GOOD_FILENAME,)
    }, name='CONFIGURATION_LAST_GOOD_FILEPATH'
)
SettingNamespaceSingleton.register_setting(
    klass=FilesystemBootstrapSetting, kwargs={
        'critical': True, 'path_parts': ('media',)
    }, name='MEDIA_ROOT'
)

# Normal settings

SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True,
        'default_value': ['127.0.0.1', 'localhost', '[::1]']
    }, name='ALLOWED_HOSTS'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='APPEND_SLASH'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='AUTH_PASSWORD_VALIDATORS'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='AUTHENTICATION_BACKENDS'
)
SettingNamespaceSingleton.register_setting(
    name='DATA_UPLOAD_MAX_MEMORY_SIZE', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': None
    }, name='DATABASES'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': False
    }, name='DEBUG'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='DEFAULT_FROM_EMAIL'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='DISALLOWED_USER_AGENTS'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='EMAIL_BACKEND'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='EMAIL_HOST'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='EMAIL_HOST_PASSWORD'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='EMAIL_HOST_USER'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='EMAIL_PORT'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='EMAIL_TIMEOUT'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='EMAIL_USE_SSL'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='EMAIL_USE_TLS'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='FILE_UPLOAD_MAX_MEMORY_SIZE'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': ['127.0.0.1']
    }, name='INTERNAL_IPS'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': 'common:home'
    }, name='LOGIN_REDIRECT_URL'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': 'authentication:login_view'
    }, name='LOGIN_URL'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': 'authentication:login_view'
    }, name='LOGOUT_REDIRECT_URL'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='LANGUAGES'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='LANGUAGE_CODE'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='SESSION_COOKIE_NAME'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='SESSION_ENGINE'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='STATIC_URL'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='STATICFILES_STORAGE'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='TIME_ZONE'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='WSGI_APPLICATION'
)

# Celery

SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': True
    }, name='CELERY_TASK_ALWAYS_EAGER'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='CELERY_BROKER_LOGIN_METHOD'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': 'memory://'
    }, name='CELERY_BROKER_URL'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='CELERY_BROKER_USE_SSL'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, name='CELERY_RESULT_BACKEND'
)

# Mayan

SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': False
    }, name='COMMON_DISABLE_LOCAL_STORAGE'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'critical': True, 'has_default': True, 'default_value': ()
    }, name='COMMON_DISABLED_APPS'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'critical': True, 'has_default': True, 'default_value': ()
    }, name='COMMON_EXTRA_APPS'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'critical': True, 'has_default': True, 'default_value': ()
    }, name='COMMON_EXTRA_APPS_PRE'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': None
    }, name='DATABASE_ENGINE'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': None
    }, name='DATABASE_NAME'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': None
    }, name='DATABASE_USER'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': None
    }, name='DATABASE_PASSWORD'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': None
    }, name='DATABASE_HOST'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': None
    }, name='DATABASE_PORT'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': 0
    }, name='DATABASE_CONN_MAX_AGE'
)
SettingNamespaceSingleton.register_setting(
    klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': False
    }, name='TESTING'
)



from django.apps import apps
from django.contrib.auth import get_user_model
from django.db import OperationalError

from mayan.apps.databases.classes import QuerysetParametersSerializer
from mayan.celery import app

from .classes import ActionExporter, EventType
from .events import event_events_cleared
from .permissions import permission_events_clear


@app.task(bind=True, ignore_result=True, retry_backoff=True)
def task_event_commit(
    self, event_id, actor_app_label=None, actor_model_name=None,
    actor_id=None, action_object_app_label=None,
    action_object_model_name=None, action_object_id=None,
    target_app_label=None, target_model_name=None, target_id=None
):
    event_type = EventType.get(id=event_id)

    try:
        if action_object_id:
            Model = apps.get_model(
                app_label=action_object_app_label,
                model_name=action_object_model_name
            )

            action_object = Model.objects.get(pk=action_object_id)
        else:
            action_object = None

        if actor_id:
            Model = apps.get_model(
                app_label=actor_app_label, model_name=actor_model_name
            )

            actor = Model.objects.get(pk=actor_id)
        else:
            actor = None

        if target_id:
            Model = apps.get_model(
                app_label=target_app_label,
                model_name=target_model_name
            )

            target = Model.objects.get(pk=target_id)
        else:
            target = None

        event_type._commit(
            action_object=action_object, actor=actor, target=target
        )
    except OperationalError as exception:
        raise self.retry(exc=exception)


@app.task(ignore_result=True)
def task_event_queryset_clear(
    decomposed_queryset, target_content_type_id=None, target_object_id=None,
    user_id=None
):
    AccessControlList = apps.get_model(
        app_label='acls', model_name='AccessControlList'
    )
    ContentType = apps.get_model(
        app_label='contenttypes', model_name='ContentType'
    )

    queryset = QuerysetParametersSerializer.rebuild(
        decomposed_queryset=decomposed_queryset
    )

    if user_id:
        user = get_user_model().objects.get(pk=user_id)
    else:
        user = None

    if target_content_type_id:
        target_content_type = ContentType.objects.get(
            pk=target_content_type_id
        )
        target = target_content_type.get_object_for_this_type(
            pk=target_object_id
        )
    else:
        target = user

    AccessControlList = apps.get_model(
        app_label='acls', model_name='AccessControlList'
    )

    if user:
        queryset = AccessControlList.objects.restrict_queryset(
            queryset=queryset,
            permission=permission_events_clear,
            user=user
        )

    commit_event = queryset.exists()

    queryset.delete()

    if commit_event:
        event_events_cleared.commit(actor=user, target=target)


@app.task(ignore_result=True)
def task_event_queryset_export(
    decomposed_queryset, organization_installation_url=None, user_id=None
):
    queryset = QuerysetParametersSerializer.rebuild(
        decomposed_queryset=decomposed_queryset
    )

    if user_id:
        user = get_user_model().objects.get(pk=user_id)
    else:
        user = None

    ActionExporter(queryset=queryset).export_to_download_file(
        organization_installation_url=organization_installation_url,
        user=user
    )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.serializers import ContentTypeSerializer
from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import MultiKwargHyperlinkedIdentityField

from ..classes import DocumentVersionModification
from ..models.document_version_models import DocumentVersion
from ..models.document_version_page_models import DocumentVersionPage


class DocumentVersionModificationSerializer(serializers.Serializer):
    id = serializers.CharField(
        label=_('ID'), read_only=True, source='backend_class_path'
    )
    label = serializers.CharField(
        label=_('Label'), read_only=True
    )
    description = serializers.CharField(
        label=_('Description'), read_only=True
    )


class DocumentVersionModificationExecuteSerializer(serializers.Serializer):
    backend_id = serializers.ChoiceField(
        choices=(), label=_('Backend ID'), help_text=_(
            'Primary key of the modification backend to execute.'
        )
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['backend_id'].choices = DocumentVersionModification.get_choices()


class DocumentVersionPageSerializer(serializers.HyperlinkedModelSerializer):
    content_type = ContentTypeSerializer(
        label=_('Content type'), read_only=True
    )
    content_type_id = serializers.IntegerField(
        help_text=_('Content type ID of the source object for the page.'),
        write_only=True
    )
    document_version_url = MultiKwargHyperlinkedIdentityField(
        label=_('Document version URL'), view_kwargs=(
            {
                'lookup_field': 'document_version.document.pk',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'document_version_id',
                'lookup_url_kwarg': 'document_version_id'
            }
        ), view_name='rest_api:documentversion-detail'
    )
    image_url = MultiKwargHyperlinkedIdentityField(
        label=_('Image URL'), view_kwargs=(
            {
                'lookup_field': 'document_version.document.pk',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'document_version_id',
                'lookup_url_kwarg': 'document_version_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'document_version_page_id'
            }
        ), view_name='rest_api:documentversionpage-image'
    )
    url = MultiKwargHyperlinkedIdentityField(
        label=_('URL'), view_kwargs=(
            {
                'lookup_field': 'document_version.document.pk',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'document_version_id',
                'lookup_url_kwarg': 'document_version_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'document_version_page_id'
            }
        ), view_name='rest_api:documentversionpage-detail'
    )

    class Meta:
        fields = (
            'content_type', 'content_type_id', 'document_version_id',
            'document_version_url', 'id', 'image_url', 'object_id',
            'page_number', 'url'
        )
        model = DocumentVersionPage
        read_only_fields = (
            'content_type', 'document_version_id', 'document_version_url',
            'id', 'image_url', 'url'
        )


class DocumentVersionSerializer(serializers.HyperlinkedModelSerializer):
    document_url = serializers.HyperlinkedIdentityField(
        label=_('Document URL'), lookup_field='document_id',
        lookup_url_kwarg='document_id', view_name='rest_api:document-detail'
    )
    export_url = MultiKwargHyperlinkedIdentityField(
        label=_('Export URL'), view_kwargs=(
            {
                'lookup_field': 'document_id',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'document_version_id'
            }
        ), view_name='rest_api:documentversion-export'
    )
    page_list_url = MultiKwargHyperlinkedIdentityField(
        label=_('Page list URL'), view_kwargs=(
            {
                'lookup_field': 'document_id',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'document_version_id'
            }
        ), view_name='rest_api:documentversionpage-list'
    )
    pages_first = DocumentVersionPageSerializer(
        label=_('Pages first'), many=False, read_only=True
    )
    url = MultiKwargHyperlinkedIdentityField(
        label=_('URL'), view_kwargs=(
            {
                'lookup_field': 'document_id',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'document_version_id'
            }
        ), view_name='rest_api:documentversion-detail'
    )

    class Meta:
        fields = (
            'active', 'comment', 'document_id', 'document_url', 'export_url',
            'id', 'page_list_url', 'pages_first', 'timestamp', 'url'
        )
        model = DocumentVersion
        read_only_fields = (
            'document_id', 'document_url', 'export_url', 'id',
            'page_list_url', 'pages_first', 'timestamp', 'url'
        )



from django.apps import apps
from django.contrib.admin.utils import help_text_for_field, label_for_field
from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.db.models import Manager
from django.db.models.query import QuerySet


class AutoHelpTextLabelFieldMixin:
    def bind(self, *args, **kwargs):
        result = super().bind(*args, **kwargs)

        try:
            model = self.root.Meta.model
        except AttributeError:
            return result
        else:
            try:
                field = model._meta.get_field(field_name=self.source)
            except FieldDoesNotExist:
                return result
            else:
                field_name = field.name

                if not self.label:
                    self.label = label_for_field(
                        model=model, name=field_name
                    )

                if not self.help_text:
                    self.help_text = help_text_for_field(
                        model=model, name=field_name
                    )

                return result


class FilteredRelatedFieldMixin:
    def __init__(self, **kwargs):
        self.source_model = kwargs.pop('source_model', None)
        self.source_permission = kwargs.pop('source_permission', None)
        self.source_queryset = kwargs.pop('source_queryset', None)
        self.source_queryset_method = kwargs.pop(
            'source_queryset_method', None
        )

        super().__init__(**kwargs)

    def get_queryset(self):
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )

        if self.source_model:
            queryset = self.source_model._meta.default_manager.all()
        elif self.source_queryset is not None:
            queryset = self.source_queryset
            if isinstance(queryset, (QuerySet, Manager)):
                # Ensure queryset is re-evaluated whenever used.
                queryset = queryset.all()
        else:
            method_name = self.source_queryset_method or 'get_{}_queryset'.format(
                self.field_name
            )
            try:
                method = getattr(self.parent, method_name)
            except AttributeError:
                raise ImproperlyConfigured(
                    'Need to provide a source_model, a '
                    'source_queryset, a source_queryset_method, or '
                    'a method named "%s".' % method_name
                )
            else:
                queryset = method()

        assert 'request' in self.context, (
            "`%s` requires the request in the serializer"
            " context. Add `context={'request': request}` when "
            "instantiating the serializer." % self.__class__.__name__
        )

        request = self.context['request']

        if self.source_permission:
            return AccessControlList.objects.restrict_queryset(
                permission=self.source_permission, queryset=queryset,
                user=request.user
            )
        else:
            return queryset



from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('checkouts', '0003_auto_20150617_0325'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='documentcheckout',
            name='user_content_type',
        ),
        migrations.RemoveField(
            model_name='documentcheckout',
            name='user_object_id',
        ),
        migrations.AlterField(
            model_name='documentcheckout',
            name='user',
            field=models.ForeignKey(
                on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL,
                verbose_name='User'
            ),
            preserve_default=True,
        ),
    ]



from mayan.apps.rest_api import generics

from .models import StoredCredential
from .permissions import (
    permission_credential_create, permission_credential_delete,
    permission_credential_edit, permission_credential_view
)
from .serializers import StoredCredentialSerializer


class APIStoredCredentialListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the credentials.
    post: Create a new credential.
    """
    mayan_object_permissions = {
        'GET': (permission_credential_view,)
    }
    mayan_view_permissions = {
        'POST': (permission_credential_create,)
    }
    serializer_class = StoredCredentialSerializer
    source_queryset = StoredCredential.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APIStoredCredentialDetailView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected credential.
    get: Return the details of the selected credential.
    patch: Edit the selected credential.
    put: Edit the selected credential.
    """
    lookup_url_kwarg = 'credential_id'
    mayan_object_permissions = {
        'DELETE': (permission_credential_delete,),
        'GET': (permission_credential_view,),
        'PATCH': (permission_credential_edit,),
        'PUT': (permission_credential_edit,)
    }
    serializer_class = StoredCredentialSerializer
    source_queryset = StoredCredential.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }



from django.contrib import admin

from .models import Role, StoredPermission


@admin.register(Role)
class RoleAdmin(admin.ModelAdmin):
    def groups_list(self, instance):
        return ','.join(
            instance.groups.values_list('name', flat=True)
        )

    def permissions_list(self, instance):
        return ','.join(
            instance.permissions.values_list('name', flat=True)
        )

    filter_horizontal = ('groups', 'permissions')
    list_display = ('label', 'permissions_list', 'groups_list')


@admin.register(StoredPermission)
class StoredPermissionAdmin(admin.ModelAdmin):
    list_display = ('namespace', 'name')
    list_display_links = list_display
    list_filter = ('namespace',)



from django.conf.urls import url

from .api_views.source_action_api_views import (
    APISourceActionDetailView, APISourceActionExecuteView,
    APISourceActionListView
)
from .api_views.source_api_views import APISourceListView, APISourceView
from .views.document_file_views import (
    DocumentFileSourceMetadataList, DocumentFileUploadView
)
from .views.document_views import DocumentUploadView
from .views.source_views import (
    SourceActionView, SourceBackendSelectionView, SourceCreateView,
    SourceDeleteView, SourceEditView, SourceListView, SourceTestView
)
from .wizards import DocumentCreateWizard

urlpatterns_documents = [
    url(
        regex=r'^sources/documents/wizard/$',
        name='document_upload_wizard', view=DocumentCreateWizard.as_view()
    ),
    url(
        regex=r'^sources/(?P<source_id>\d+)/documents/upload/$',
        name='document_upload', view=DocumentUploadView.as_view()
    ),
    url(
        regex=r'^sources/any/documents/upload/$',
        name='document_upload', view=DocumentUploadView.as_view()
    )
]

urlpatterns_document_files = [
    url(
        regex=r'^document_files/(?P<document_file_id>\d+)/source_metadata/$',
        name='document_file_source_metadata_list',
        view=DocumentFileSourceMetadataList.as_view()
    ),
    url(
        regex=r'^sources/(?P<source_id>\d+)/documents/(?P<document_id>\d+)/files/upload/$',
        name='document_file_upload', view=DocumentFileUploadView.as_view()
    ),
    url(
        regex=r'^sources/any/documents/(?P<document_id>\d+)/files/upload/$',
        name='document_file_upload', view=DocumentFileUploadView.as_view()
    )
]

urlpatterns_sources = [

    url(
        regex=r'^sources/$', name='source_list',
        view=SourceListView.as_view()
    ),
    url(
        regex=r'^sources/backend/selection/$',
        name='source_backend_selection',
        view=SourceBackendSelectionView.as_view()
    ),
    url(
        regex=r'^sources/(?P<backend_path>[a-zA-Z0-9_.]+)/create/$',
        name='source_create', view=SourceCreateView.as_view()
    ),
    url(
        regex=r'^sources/(?P<source_id>\d+)/actions/(?P<action_name>[a-zA-Z0-9_.]+)/$',
        name='source_action', view=SourceActionView.as_view()
    ),
    url(
        regex=r'^sources/(?P<source_id>\d+)/delete/$',
        name='source_delete', view=SourceDeleteView.as_view()
    ),
    url(
        regex=r'^sources/(?P<source_id>\d+)/edit/$', name='source_edit',
        view=SourceEditView.as_view()
    ),
    url(
        regex=r'^sources/(?P<source_id>\d+)/test/$',
        name='source_test', view=SourceTestView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_documents)
urlpatterns.extend(urlpatterns_document_files)
urlpatterns.extend(urlpatterns_sources)

api_urls = [
    url(
        regex=r'^sources/$', name='source-list',
        view=APISourceListView.as_view()
    ),
    url(
        regex=r'^sources/(?P<source_id>[0-9]+)/$',
        name='source-detail', view=APISourceView.as_view()
    ),
    url(
        regex=r'^sources/(?P<source_id>[0-9]+)/actions/$',
        name='source_action-list', view=APISourceActionListView.as_view()
    ),
    url(
        regex=r'^sources/(?P<source_id>[0-9]+)/actions/(?P<action_name>[a-zA-Z0-9_.]+)/$',
        name='source_action-detail', view=APISourceActionDetailView.as_view()
    ),
    url(
        regex=r'^sources/(?P<source_id>[0-9]+)/actions/(?P<action_name>[a-zA-Z0-9_.]+)/execute/$',
        name='source_action-execute', view=APISourceActionExecuteView.as_view()
    )
]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import (
    icon_credential_backend_selection, icon_credential_delete,
    icon_credential_edit, icon_credential_list, icon_credential_setup
)
from .permissions import (
    permission_credential_create, permission_credential_delete,
    permission_credential_edit, permission_credential_view
)

link_credential_backend_selection = Link(
    icon=icon_credential_backend_selection,
    permissions=(permission_credential_create,),
    text=_('Create credential'),
    view='credentials:stored_credential_backend_selection'
)
link_credential_delete = Link(
    args='resolved_object.pk',
    icon=icon_credential_delete,
    permissions=(permission_credential_delete,),
    tags='dangerous', text=_('Delete'),
    view='credentials:stored_credential_delete',
)
link_credential_edit = Link(
    args='object.pk',
    icon=icon_credential_edit,
    permissions=(permission_credential_edit,),
    text=_('Edit'), view='credentials:stored_credential_edit'
)
link_credential_list = Link(
    icon=icon_credential_list, text=_('Credential list'),
    view='credentials:stored_credential_list'
)
link_credential_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='credentials', model_name='StoredCredential',
        object_permission=permission_credential_view,
        view_permission=permission_credential_create,
    ), icon=icon_credential_setup,
    text=_('Credentials'), view='credentials:stored_credential_list'
)



from django.core import mail

from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import (
    event_email_sent, event_profile_created, event_profile_edited
)
from ..models import UserMailer
from ..permissions import (
    permission_user_mailer_create, permission_user_mailer_delete,
    permission_user_mailer_edit, permission_user_mailer_use,
    permission_user_mailer_view
)

from .literals import (
    TEST_EMAIL_ADDRESS, TEST_EMAIL_FROM_ADDRESS,
    TEST_RECIPIENTS_MULTIPLE_COMMA, TEST_RECIPIENTS_MULTIPLE_COMMA_RESULT,
    TEST_RECIPIENTS_MULTIPLE_MIXED, TEST_RECIPIENTS_MULTIPLE_MIXED_RESULT,
    TEST_RECIPIENTS_MULTIPLE_SEMICOLON,
    TEST_RECIPIENTS_MULTIPLE_SEMICOLON_RESULT
)
from .mixins import MailerViewTestMixin


class MailerViewTestCase(MailerViewTestMixin, GenericViewTestCase):
    def test_user_mailer_create_view_no_permission(self):
        self.grant_permission(permission=permission_user_mailer_view)

        self._clear_events()

        response = self._request_test_user_mailer_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(
            UserMailer.objects.count(), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_mailer_create_view_with_permissions(self):
        self.grant_permission(permission=permission_user_mailer_create)
        self.grant_permission(permission=permission_user_mailer_view)

        self._clear_events()

        response = self._request_test_user_mailer_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            UserMailer.objects.count(), 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_user_mailer)
        self.assertEqual(events[0].verb, event_profile_created.id)

    def test_user_mailer_delete_view_no_permission(self):
        self._create_test_user_mailer()

        self._clear_events()

        response = self._request_test_user_mailer_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertQuerysetEqual(
            UserMailer.objects.all(), (
                repr(self._test_user_mailer),
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_mailer_delete_view_with_access(self):
        self._create_test_user_mailer()

        self.grant_access(
            obj=self._test_user_mailer,
            permission=permission_user_mailer_delete
        )

        self._clear_events()

        response = self._request_test_user_mailer_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            UserMailer.objects.count(), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_mailer_edit_view_no_permission(self):
        self._create_test_user_mailer()

        test_profile_label = self._test_user_mailer.label

        self._clear_events()

        response = self._request_test_user_mailer_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_user_mailer.refresh_from_db()
        self.assertEqual(
            self._test_user_mailer.label, test_profile_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_mailer_edit_view_with_access(self):
        self._create_test_user_mailer()

        self.grant_access(
            obj=self._test_user_mailer,
            permission=permission_user_mailer_edit
        )

        test_profile_label = self._test_user_mailer.label

        self._clear_events()

        response = self._request_test_user_mailer_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_user_mailer.refresh_from_db()
        self.assertNotEqual(
            self._test_user_mailer.label, test_profile_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_user_mailer)
        self.assertEqual(events[0].verb, event_profile_edited.id)

    def test_user_mailer_list_view_no_permission(self):
        self._create_test_user_mailer()

        self._clear_events()

        response = self._request_test_user_mailer_list_view()
        self.assertNotContains(
            response=response, status_code=200,
            text=self._test_user_mailer.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_mailer_list_view_with_access(self):
        self._create_test_user_mailer()

        self.grant_access(
            obj=self._test_user_mailer,
            permission=permission_user_mailer_view
        )

        self._clear_events()

        response = self._request_test_user_mailer_list_view()
        self.assertContains(
            response=response, status_code=200,
            text=self._test_user_mailer.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_mailer_list_bad_data_view_with_access(self):
        self._silence_logger(name='mayan.apps.backends.model_mixins')

        self._create_test_user_mailer()
        self._test_user_mailer.backend_path = 'bad.backend.path'
        self._test_user_mailer.backend_data = '{"bad_field": "bad_data"}'
        self._test_user_mailer.save()

        self.grant_access(
            obj=self._test_user_mailer,
            permission=permission_user_mailer_view
        )

        self._clear_events()

        response = self._request_test_user_mailer_list_view()
        self.assertContains(
            response=response, status_code=200,
            text=self._test_user_mailer.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_mailer_test_view_no_permission(self):
        self._create_test_user_mailer()

        self._clear_events()

        response = self._request_test_user_mailer_test_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_mailer_test_view_with_access(self):
        self._create_test_user_mailer()

        self.grant_access(
            obj=self._test_user_mailer,
            permission=permission_user_mailer_use
        )

        self._clear_events()

        response = self._request_test_user_mailer_test_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            len(mail.outbox), 1
        )
        self.assertEqual(
            mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS
        )
        self.assertEqual(
            mail.outbox[0].to, [TEST_EMAIL_ADDRESS]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_send_multiple_recipients_comma(self):
        self._create_test_user_mailer()

        self.grant_access(
            obj=self._test_user_mailer,
            permission=permission_user_mailer_use
        )

        self._test_email_address = TEST_RECIPIENTS_MULTIPLE_COMMA

        self._clear_events()

        response = self._request_test_user_mailer_test_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            len(mail.outbox), 1
        )
        self.assertEqual(
            mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS
        )
        self.assertEqual(
            mail.outbox[0].to, TEST_RECIPIENTS_MULTIPLE_COMMA_RESULT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_send_multiple_recipients_mixed(self):
        self._create_test_user_mailer()

        self.grant_access(
            obj=self._test_user_mailer,
            permission=permission_user_mailer_use
        )

        self._test_email_address = TEST_RECIPIENTS_MULTIPLE_MIXED

        self._clear_events()

        response = self._request_test_user_mailer_test_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            len(mail.outbox), 1
        )
        self.assertEqual(
            mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS
        )
        self.assertEqual(
            mail.outbox[0].to, TEST_RECIPIENTS_MULTIPLE_MIXED_RESULT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_send_multiple_recipients_semicolon(self):
        self._create_test_user_mailer()

        self.grant_access(
            obj=self._test_user_mailer, permission=permission_user_mailer_use
        )

        self._test_email_address = TEST_RECIPIENTS_MULTIPLE_SEMICOLON

        self._clear_events()

        response = self._request_test_user_mailer_test_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            len(mail.outbox), 1
        )
        self.assertEqual(
            mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS
        )
        self.assertEqual(
            mail.outbox[0].to, TEST_RECIPIENTS_MULTIPLE_SEMICOLON_RESULT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)



from rest_framework.filters import BaseFilterBackend, OrderingFilter

from mayan.apps.acls.models import AccessControlList


class MayanObjectPermissionsFilter(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        permission = self.get_mayan_object_permissions(
            request=request, view=view
        )

        if permission:
            return AccessControlList.objects.restrict_queryset(
                queryset=queryset, permission=permission,
                user=request.user
            )
        else:
            return queryset

    def get_mayan_object_permissions(self, request, view):
        try:
            method_get_mayan_object_permissions = getattr(
                view, 'get_mayan_object_permissions'
            )
        except AttributeError:
            return getattr(
                view, 'mayan_object_permissions', {}
            ).get(
                request.method, (None,)
            )[0]
        else:
            return method_get_mayan_object_permissions()


class MayanSortingFilter(OrderingFilter):
    ordering_param = '_ordering'

    def get_default_valid_fields(self, queryset, view, context=None):
        return ()



class RecentlyAccessedDocumentAPIViewTestMixin:
    def _request_recently_accessed_document_list_api_view(self):
        return self.get(
            viewname='rest_api:recentlyaccesseddocument-list'
        )


class RecentlyAccessedDocumentViewTestMixin:
    def _request_test_recently_accessed_document_list_view(self):
        return self.get(viewname='documents:document_recently_accessed_list')



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('common', '0004_delete_anonymoususersingleton'),
    ]

    operations = [
        migrations.RenameField(
            model_name='shareduploadedfile',
            old_name='datatime',
            new_name='datetime',
        ),
    ]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.search import search_model_document
from mayan.apps.dynamic_search.search_models import SearchModel

from .permissions import permission_tag_view

# Document

search_model_document.add_model_field(
    field='tags__label', label=_('Tag label')
)
search_model_document.add_model_field(
    field='tags__color', label=_('Tag color')
)

# Tag

search_model_tag = SearchModel(
    app_label='tags', model_name='Tag',
    permission=permission_tag_view,
    serializer_path='mayan.apps.tags.serializers.TagSerializer'
)
search_model_tag.add_model_field(field='label')
search_model_tag.add_model_field(field='color')

search_model_tag.add_model_field(
    field='documents__document_type__label', label=_('Document type')
)
search_model_tag.add_model_field(
    field='documents__label', label=_('Document label')
)
search_model_tag.add_model_field(
    field='documents__description', label=_('Document description')
)
search_model_tag.add_model_field(
    field='documents__uuid', label=_('Document UUID')
)

search_model_tag.add_model_field(
    field='documents__files__checksum', label=_('Document file checksum')
)
search_model_tag.add_model_field(
    field='documents__files__mimetype', label=_('Document file MIME type')
)



from django import forms
from django.utils.translation import ugettext_lazy as _

from mayan.apps.backends.forms import FormDynamicModelBackend
from mayan.apps.templating.fields import ModelTemplateField

from ..classes import WorkflowAction
from ..models import (
    WorkflowInstance, WorkflowState, WorkflowStateAction,
    WorkflowStateEscalation
)


class WorkflowTemplateStateActionDynamicForm(FormDynamicModelBackend):
    class Meta:
        fields = ('label', 'when', 'enabled', 'condition', 'backend_data')
        model = WorkflowStateAction
        widgets = {'backend_data': forms.widgets.HiddenInput}

    def __init__(self, request, user=None, *args, **kwargs):
        self.request = request
        self.user = user
        result = super().__init__(*args, **kwargs)

        self.fields['condition'] = ModelTemplateField(
            initial_help_text=self.fields['condition'].help_text,
            label=self.fields['condition'].label, model=WorkflowInstance,
            model_variable='workflow_instance', required=False
        )

        return result


class WorkflowTemplateStateActionSelectionForm(forms.Form):
    klass = forms.ChoiceField(
        choices=(), help_text=_('The action type for this action entry.'),
        label=_('Action'), widget=forms.Select(
            attrs={'class': 'select2'}
        )
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields['klass'].choices = WorkflowAction.get_choices()


class WorkflowTemplateStateEscalationForm(forms.ModelForm):
    def __init__(self, workflow_template_state, *args, **kwargs):
        self.workflow_template_state = workflow_template_state
        super().__init__(*args, **kwargs)

        queryset_transitions = self.workflow_template_state.workflow.transitions.filter(
            origin_state=self.workflow_template_state
        )

        self.fields['transition'].queryset = queryset_transitions

        self.fields['condition'] = ModelTemplateField(
            initial_help_text=self.fields['condition'].help_text,
            label=self.fields['condition'].label, model=WorkflowInstance,
            model_variable='workflow_instance', required=False
        )

    class Meta:
        fields = (
            'enabled', 'transition', 'priority', 'unit', 'amount',
            'condition'
        )
        model = WorkflowStateEscalation


class WorkflowTemplateStateForm(forms.ModelForm):
    class Meta:
        fields = ('initial', 'label', 'completion')
        model = WorkflowState



from django.utils.translation import ugettext_lazy as _

from mayan.apps.document_states.classes import WorkflowAction

from .literals import MODEL_SEND_FUNCTION_DOTTED_PATH
from .workflow_action_mixins import ObjectEmailActionMixin


class DocumentEmailAction(ObjectEmailActionMixin, WorkflowAction):
    label = _('Send document via email')
    previous_dotted_paths = (
        'mayan.apps.mailer.workflow_actions.EmailAction',
    )

    @classmethod
    def get_form_fields(cls, *args, **kwargs):
        fields = super().get_form_fields(*args, **kwargs)

        fields.update(
            {
                'attachment': {
                    'label': _('Attachment'),
                    'class': 'django.forms.BooleanField', 'default': False,
                    'help_text': _(
                        'Attach the exported document version to the email.'
                    ),
                    'required': False
                }
            }
        )

        return fields

    def get_execute_data(self, context):
        result = super().get_execute_data(context=context)
        document = self.get_object(context=context)

        if self.kwargs.get('attachment', False):
            if document.version_active:
                # Document must have a version active in order to be able
                # to export and attach.
                obj = document.version_active
                result.update(
                    {
                        'as_attachment': True,
                        'obj': obj
                    }
                )
                result.update(
                    MODEL_SEND_FUNCTION_DOTTED_PATH.get(
                        obj._meta.model, {}
                    )
                )

        return result

    def get_object(self, context):
        return context['workflow_instance'].document



import bleach

from django.db import migrations


def code_sanitize_themes(apps, schema_editor):
    """
    Sanitize all theme stylesheets.
    """
    Theme = apps.get_model(
        app_label='appearance', model_name='Theme'
    )

    for theme in Theme.objects.all():
        theme.stylesheet = bleach.clean(
            text=theme.stylesheet, tags=('style',)
        )
        theme.save()


class Migration(migrations.Migration):
    dependencies = [
        ('appearance', '0002_add_theme_for_existing_users')
    ]

    operations = [
        migrations.RunPython(
            code=code_sanitize_themes,
            reverse_code=migrations.RunPython.noop
        )
    ]



from django.db import migrations
import django.db.models.deletion
import mptt.fields


class Migration(migrations.Migration):
    dependencies = [
        ('document_indexing', '0024_auto_20210822_0052')
    ]

    operations = [
        migrations.AlterField(
            model_name='indextemplatenode',
            name='parent',
            field=mptt.fields.TreeForeignKey(
                blank=True,
                help_text='Parent index template node of this node.',
                null=True, on_delete=django.db.models.deletion.CASCADE,
                related_name='children',
                to='document_indexing.IndexTemplateNode'
            )
        )
    ]



from mayan.apps.documents.permissions import (
    permission_document_create, permission_document_file_new
)
from mayan.apps.source_compressed.source_backend_actions.mixins import SourceBackendActionMixinCompressedInteractive
from mayan.apps.sources.source_backend_actions.base import SourceBackendAction
from mayan.apps.sources.source_backend_actions.mixins.document_file_mixins import SourceBackendActionMixinDocumentFileUploadInteractive
from mayan.apps.sources.source_backend_actions.mixins.document_mixins import (
    SourceBackendActionMixinDocumentInteractive,
    SourceBackendActionMixinDocumentUploadInteractive
)
from mayan.apps.sources.source_backend_actions.mixins.document_type_mixins import SourceBackendActionMixinDocumentTypeInteractive
from mayan.apps.sources.source_backend_actions.mixins.immediate_mode_mixins import SourceBackendActionMixinImmediateMode

from .callback_mixins import (
    SourceBackendActionMixinCallbackDocumentFileUploadInteractive,
    SourceBackendActionMixinCallbackDocumentUploadInteractive
)
from .file_mixins import SourceBackendActionMixinFileUser


class SourceBackendActionInteractiveBase(
    SourceBackendActionMixinFileUser, SourceBackendAction
):
    """
    Base class for all interactive source actions.
    """


class SourceBackendActionInteractiveDocumentFileUpload(
    SourceBackendActionMixinDocumentFileUploadInteractive,
    SourceBackendActionMixinDocumentInteractive,
    SourceBackendActionMixinCallbackDocumentFileUploadInteractive,
    SourceBackendActionInteractiveBase
):
    name = 'document_file_upload'
    permission = permission_document_file_new


class SourceBackendActionInteractiveDocumentUpload(
    SourceBackendActionMixinDocumentUploadInteractive,
    SourceBackendActionMixinCompressedInteractive,
    SourceBackendActionMixinImmediateMode,
    SourceBackendActionMixinCallbackDocumentUploadInteractive,
    SourceBackendActionMixinDocumentTypeInteractive,
    SourceBackendActionInteractiveBase
):
    name = 'document_upload'
    permission = permission_document_create



from furl import furl

from django.conf import settings
from django.test import override_settings
from django.urls import reverse
from django.utils.http import urlunquote_plus

from mayan.apps.authentication.classes import AuthenticationBackend
from mayan.apps.authentication.events import event_user_logged_in
from mayan.apps.authentication.tests.mixins import LoginViewTestMixin
from mayan.apps.smart_settings.classes import SettingNamespace
from mayan.apps.testing.tests.base import GenericViewTestCase
from mayan.apps.user_management.events import event_user_edited
from mayan.apps.views.http import URL

from .literals import (
    PATH_AUTHENTICATION_BACKEND_EMAIL_OTP,
    PATH_AUTHENTICATION_BACKEND_USERNAME_OTP
)
from .mixins import AuthenticationOTPTestMixin


class LoginOTPTestCase(
    AuthenticationOTPTestMixin, LoginViewTestMixin, GenericViewTestCase
):
    authenticated_url = reverse(viewname='common:home')
    authentication_url = urlunquote_plus(
        furl(
            path=reverse(settings.LOGIN_URL), args={
                'next': authenticated_url
            }
        ).tostr()
    )
    auto_login_user = False
    create_test_case_super_user = True

    def setUp(self):
        super().setUp()
        SettingNamespace.invalidate_cache_all()

    @override_settings(AUTHENTICATION_BACKEND=PATH_AUTHENTICATION_BACKEND_EMAIL_OTP)
    def test_login_view_with_email_no_otp(self):
        AuthenticationBackend.cls_initialize()

        self._clear_events()

        response = self._request_login_view_with_email()
        self.assertEqual(response.status_code, 302)

        response = self._request_multi_factor_authentication_view()
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_super_user)
        self.assertEqual(events[0].target, self._test_case_super_user)
        self.assertEqual(events[0].verb, event_user_edited.id)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_super_user)
        self.assertEqual(events[1].target, self._test_case_super_user)
        self.assertEqual(events[1].verb, event_user_logged_in.id)

    @override_settings(AUTHENTICATION_BACKEND=PATH_AUTHENTICATION_BACKEND_EMAIL_OTP)
    def test_login_view_with_email_with_otp(self):
        AuthenticationBackend.cls_initialize()

        self._enable_test_otp()

        self._clear_events()

        response = self._request_login_view_with_email()
        self.assertEqual(response.status_code, 302)

        response = self._request_multi_factor_authentication_view(
            data={
                '0-token': self._test_token
            }
        )
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_super_user)
        self.assertEqual(events[0].target, self._test_case_super_user)
        self.assertEqual(events[0].verb, event_user_edited.id)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_super_user)
        self.assertEqual(events[1].target, self._test_case_super_user)
        self.assertEqual(events[1].verb, event_user_logged_in.id)

    @override_settings(AUTHENTICATION_BACKEND=PATH_AUTHENTICATION_BACKEND_USERNAME_OTP)
    def test_login_view_with_username_no_otp(self):
        AuthenticationBackend.cls_initialize()

        self._clear_events()

        response = self._request_login_view_with_username()
        self.assertEqual(response.status_code, 302)

        response = self._request_multi_factor_authentication_view()
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_super_user)
        self.assertEqual(events[0].target, self._test_case_super_user)
        self.assertEqual(events[0].verb, event_user_edited.id)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_super_user)
        self.assertEqual(events[1].target, self._test_case_super_user)
        self.assertEqual(events[1].verb, event_user_logged_in.id)

    @override_settings(AUTHENTICATION_BACKEND=PATH_AUTHENTICATION_BACKEND_USERNAME_OTP)
    def test_login_view_with_username_with_otp(self):
        AuthenticationBackend.cls_initialize()

        self._enable_test_otp()

        self._clear_events()

        response = self._request_login_view_with_username()
        self.assertEqual(response.status_code, 302)

        response = self._request_multi_factor_authentication_view(
            data={
                '0-token': self._test_token
            }
        )
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_super_user)
        self.assertEqual(events[0].target, self._test_case_super_user)
        self.assertEqual(events[0].verb, event_user_edited.id)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_super_user)
        self.assertEqual(events[1].target, self._test_case_super_user)
        self.assertEqual(events[1].verb, event_user_logged_in.id)

    @override_settings(AUTHENTICATION_BACKEND=PATH_AUTHENTICATION_BACKEND_USERNAME_OTP)
    def test_login_view_redirect_with_username_and_otp(self):
        AuthenticationBackend.cls_initialize()

        TEST_REDIRECT_URL = reverse(viewname='common:about_view')
        self._enable_test_otp()

        self._clear_events()

        response = self._request_login_view_with_username(
            follow=True, query={'next': TEST_REDIRECT_URL}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.redirect_chain, [
                (
                    URL(
                        path=reverse(
                            viewname='authentication:multi_factor_authentication_view'
                        ), query={'next': TEST_REDIRECT_URL}
                    ).to_string(), 302
                )
            ]
        )

        response = self._request_multi_factor_authentication_view(
            data={
                '0-token': self._test_token
            }, follow=True, query={'next': TEST_REDIRECT_URL}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.redirect_chain, [
                (TEST_REDIRECT_URL, 302)
            ]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_super_user)
        self.assertEqual(events[0].target, self._test_case_super_user)
        self.assertEqual(events[0].verb, event_user_edited.id)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_super_user)
        self.assertEqual(events[1].target, self._test_case_super_user)
        self.assertEqual(events[1].verb, event_user_logged_in.id)


class UserOTPDataViewTestCase(
    AuthenticationOTPTestMixin, GenericViewTestCase
):
    @override_settings(AUTHENTICATION_BACKEND=PATH_AUTHENTICATION_BACKEND_USERNAME_OTP)
    def test_user_otp_data_verify_token_view(self):
        response = self.get(viewname='authentication_otp:otp_verify')
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.urls import reverse

from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.documents.tests.mixins.document_file_mixins import DocumentFileTestMixin

from ..links import link_document_file_download_quick
from ..permissions import permission_document_file_download


class DocumentsLinksTestCase(
    DocumentFileTestMixin, GenericDocumentViewTestCase
):
    def test_document_file_download_link_no_permission(self):
        self.add_test_view(test_object=self._test_document.file_latest)
        context = self.get_test_view()
        resolved_link = link_document_file_download_quick.resolve(
            context=context
        )

        self.assertEqual(resolved_link, None)

    def test_document_file_download_link_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_download
        )

        self.add_test_view(test_object=self._test_document.file_latest)
        context = self.get_test_view()
        resolved_link = link_document_file_download_quick.resolve(
            context=context
        )

        self.assertNotEqual(resolved_link, None)
        self.assertEqual(
            resolved_link.url,
            reverse(
                viewname=link_document_file_download_quick.view,
                args=(self._test_document.file_latest.pk,)
            )
        )



from django.apps import apps

from mayan.apps.converter.settings import setting_image_generation_timeout

from .events import event_ocr_document_version_submitted
from .literals import TASK_DOCUMENT_VERSION_PAGE_OCR_TIMEOUT
from .tasks import task_document_version_ocr_process


def method_document_ocr_content(self):
    version_active = self.version_active

    if version_active:
        return version_active.ocr_content()
    else:
        return ()


def method_document_ocr_submit(self, user=None):
    version_active = self.version_active
    # Don't error out if document has no version.
    if version_active:
        version_active.submit_for_ocr(user=user)


def method_document_version_ocr_content(self):
    DocumentVersionPageOCRContent = apps.get_model(
        app_label='ocr', model_name='DocumentVersionPageOCRContent'
    )

    for page in self.pages.all():
        try:
            page_content = page.ocr_content.content
        except DocumentVersionPageOCRContent.DoesNotExist:
            """Not critical, just ignore and go to next page."""
        else:
            yield page_content


def method_document_version_ocr_submit(self, user=None):
    if user:
        user_id = user.pk
    else:
        user_id = None

    # Timeout calculation logic:
    # Total timeout for the document version task should be the
    # file page render timeout + the version page timeout + version page OCR
    # timeout x the total version pages. This decreases the probability of
    # the OCR task getting killed before the document file and version page
    # rendering finishes.

    event_ocr_document_version_submitted.commit(
        action_object=self.document, actor=user, target=self
    )

    task_document_version_ocr_process.apply_async(
        kwargs={
            'document_version_id': self.pk, 'user_id': user_id
        }, timeout=(
            TASK_DOCUMENT_VERSION_PAGE_OCR_TIMEOUT + setting_image_generation_timeout.value * 2
        ) * self.pages.count()
    )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(
    label=_('Mailing'), name='mailing'
)

# Mailing profile

permission_user_mailer_create = namespace.add_permission(
    label=_('Create a mailing profile'), name='user_mailer_create'
)
permission_user_mailer_delete = namespace.add_permission(
    label=_('Delete a mailing profile'), name='user_mailer_delete'
)
permission_user_mailer_edit = namespace.add_permission(
    label=_('Edit a mailing profile'), name='user_mailer_edit'
)
permission_user_mailer_view = namespace.add_permission(
    label=_('View a mailing profile'), name='user_mailer_view'
)
permission_user_mailer_use = namespace.add_permission(
    label=_('Use a mailing profile'), name='user_mailer_use'
)

# Document

permission_send_document_link = namespace.add_permission(
    label=_('Send document link via email'), name='mail_link'
)

# Document file

permission_send_document_file_attachment = namespace.add_permission(
    label=_('Send document file via email'),
    name='mail_document_file_attachment'
)
permission_send_document_file_link = namespace.add_permission(
    label=_('Send document file link via email'),
    name='mail_document_file_link'
)

# Document version

permission_send_document_version_attachment = namespace.add_permission(
    label=_('Send document version via email'),
    name='mail_document_version_attachment'
)
permission_send_document_version_link = namespace.add_permission(
    label=_('Send document version link via email'),
    name='mail_document_version_link'
)



from mayan.apps.appearance.classes import Icon

# Document

icon_document_link_send_single = Icon(
    driver_name='fontawesome', symbol='external-link-alt'
)
icon_document_link_send_multiple = Icon(
    driver_name='fontawesome', symbol='external-link-alt'
)

# Document file

icon_document_file_attachment_send_single = Icon(
    driver_name='fontawesome', symbol='paperclip'
)
icon_document_file_attachment_send_multiple = Icon(
    driver_name='fontawesome', symbol='external-link-alt'
)
icon_document_file_link_send_single = Icon(
    driver_name='fontawesome', symbol='paperclip'
)
icon_document_file_link_send_multiple = Icon(
    driver_name='fontawesome', symbol='external-link-alt'
)

# Document version

icon_document_version_attachment_send_multiple = Icon(
    driver_name='fontawesome', symbol='paperclip'
)
icon_document_version_attachment_send_single = Icon(
    driver_name='fontawesome', symbol='paperclip'
)
icon_document_version_link_send_multiple = Icon(
    driver_name='fontawesome', symbol='external-link-alt'
)
icon_document_version_link_send_single = Icon(
    driver_name='fontawesome', symbol='external-link-alt'
)

# Mailing profile

icon_user_mailer_backend_select = Icon(
    driver_name='fontawesome-dual', primary_symbol='envelope',
    secondary_symbol='plus'
)
icon_user_mailer_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='envelope',
    secondary_symbol='plus'
)
icon_user_mailer_delete = Icon(driver_name='fontawesome', symbol='times')
icon_user_mailer_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_user_mailer_list = Icon(driver_name='fontawesome', symbol='envelope')
icon_user_mailer_test = Icon(driver_name='fontawesome', symbol='play')



from django.conf import settings

from mayan.apps.authentication.authentication_backends import (
    AuthenticationBackendModelEmailPassword,
    AuthenticationBackendModelUsernamePassword
)
from mayan.apps.common.utils import get_class_full_name

from .django_authentication_backends import (
    DjangoAuthenticationBackendEmailMultiFactorOTP,
    DjangoAuthenticationBackendUsernameMultiFactorOTP
)
from .forms import AuthenticationFormTOTP


class AuthenticationBackendTOTPMixin:
    def __init__(self, **kwargs):
        self.form_list += (AuthenticationFormTOTP,)

        super().__init__(**kwargs)


class AuthenticationBackendModelEmailPasswordTOTP(
    AuthenticationBackendTOTPMixin, AuthenticationBackendModelEmailPassword
):
    """
    Same backend as AuthenticationBackendModelEmailPassword but with
    an additional form for an TOTP token.
    """
    def initialize(self):
        settings.AUTHENTICATION_BACKENDS = (
            get_class_full_name(klass=DjangoAuthenticationBackendEmailMultiFactorOTP),
        )


class AuthenticationBackendModelUsernamePasswordTOTP(
    AuthenticationBackendTOTPMixin,
    AuthenticationBackendModelUsernamePassword
):
    """
    Same backend as AuthenticationBackendModelUsernamePassword but with
    an additional form for an TOTP token.
    """
    def initialize(self):
        settings.AUTHENTICATION_BACKENDS = (
            get_class_full_name(klass=DjangoAuthenticationBackendUsernameMultiFactorOTP),
        )



import hashlib

from django.core import serializers
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.serialization import yaml_load
from mayan.apps.templating.classes import Template


class WorkflowTransitionBusinessLogicMixin:
    def evaluate_condition(self, workflow_instance):
        if self.has_condition():
            return Template(template_string=self.condition).render(
                context={'workflow_instance': workflow_instance}
            ).strip()
        else:
            return True

    def get_field_display(self):
        field_list = [
            str(field) for field in self.fields.all()
        ]
        field_list.sort()

        return ', '.join(field_list)

    get_field_display.short_description = _('Fields')

    def get_hash(self):
        result = hashlib.sha256(
            string=serializers.serialize(
                format='json', queryset=(self,)
            ).encode()
        )
        for trigger_event in self.trigger_events.all():
            result.update(
                trigger_event.get_hash().encode()
            )

        for field in self.fields.all():
            result.update(
                field.get_hash().encode()
            )

        return result.hexdigest()

    def has_condition(self):
        return self.condition.strip()
    has_condition.help_text = _(
        'The transition will be available, depending on the condition '
        'return value.'
    )
    has_condition.short_description = _('Has a condition?')


class WorkflowTransitionFieldBusinessLogicMixin:
    def get_hash(self):
        return hashlib.sha256(
            string=serializers.serialize(
                format='json', queryset=(self,)
            ).encode()
        ).hexdigest()

    def get_widget_kwargs(self):
        return yaml_load(
            stream=self.widget_kwargs or '{}'
        )


class WorkflowTransitionTriggerEventBusinessLogicMixin:
    def get_hash(self):
        return hashlib.sha256(
            string=serializers.serialize(
                format='json', queryset=(self,)
            ).encode()
        ).hexdigest()



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(
    label=_('User management'), name='user_management'
)

permission_group_create = namespace.add_permission(
    label=_('Create new groups'), name='group_create'
)
permission_group_delete = namespace.add_permission(
    label=_('Delete existing groups'), name='group_delete'
)
permission_group_edit = namespace.add_permission(
    label=_('Edit existing groups'), name='group_edit'
)
permission_group_view = namespace.add_permission(
    label=_('View existing groups'), name='group_view'
)
permission_user_create = namespace.add_permission(
    label=_('Create new users'), name='user_create'
)
permission_user_delete = namespace.add_permission(
    label=_('Delete existing users'), name='user_delete'
)
permission_user_edit = namespace.add_permission(
    label=_('Edit existing users'), name='user_edit'
)
permission_user_view = namespace.add_permission(
    label=_('View existing users'), name='user_view'
)



from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.search import search_model_document
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.dynamic_search.tests.mixins.base import SearchTestMixin

from .mixins import TagTestMixin


class DocumentTagSearchTestCase(
    TagTestMixin, SearchTestMixin, GenericDocumentViewTestCase
):
    _test_search_model = search_model_document
    _test_tag_add_test_document = True
    auto_create_test_tag = True

    def test_search_model_document_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'tags__label': self._test_tag.label
            }
        )
        self.assertTrue(self._test_document not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_search_model_document_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'tags__label': self._test_tag.label
            }
        )
        self.assertTrue(self._test_document in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_search_model_document_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'tags__label': self._test_tag.label
            }
        )
        self.assertTrue(self._test_document not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import logging

from celery.backends.base import DisabledBackend

from django.apps import apps
from django.conf import settings
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import menu_list_facet, menu_tools, menu_return
from mayan.apps.common.signals import signal_perform_upgrade
from mayan.apps.navigation.classes import SourceColumn
from mayan.apps.views.column_widgets import TwoStateWidget
from mayan.celery import app as celery_app

from .classes import CeleryQueue, TaskType, Task, Worker
from .handlers import handler_perform_upgrade
from .links import (
    link_queue_task_type_list, link_worker_list, link_worker_queue_list
)
from .literals import TEST_CELERY_RESULT_KEY, TEST_CELERY_RESULT_VALUE
from .methods import factory_method_periodic_task_save

logger = logging.getLogger(name=__name__)


class TaskManagerApp(MayanAppConfig):
    app_namespace = 'task_manager'
    app_url = 'task_manager'
    has_tests = True
    name = 'mayan.apps.task_manager'
    verbose_name = _('Task manager')

    def check_broker_connectivity(self):
        connection = celery_app.connection()

        logger.debug('Starting Celery broker connectivity test')
        try:
            connection.ensure_connection(
                interval_step=0, interval_max=0, interval_start=0,
                timeout=0.1
            )
        except Exception as exception:
            print(
                'Failed to connect to the Celery broker at {}; {}'.format(
                    connection.as_uri(), exception
                )
            )
            raise
        else:
            connection.release()

    def check_results_backend_connectivity(self):
        backend = celery_app.backend

        if not isinstance(backend, DisabledBackend):
            retry_policy = backend.retry_policy

            backend.retry_policy = {
                'max_retries': 0, 'interval_start': 0, 'interval_step': 1,
                'interval_max': 1
            }

            logger.debug('Starting Celery result backend connectivity test')
            try:
                backend.set(
                    key=TEST_CELERY_RESULT_KEY,
                    value=TEST_CELERY_RESULT_VALUE
                )
            except Exception as exception:
                print(
                    'Failed to connect to the Celery result backend at {}; {}'.format(
                        backend.as_uri(), exception
                    )
                )
                raise
            else:
                backend.delete(key=TEST_CELERY_RESULT_KEY)
                backend.retry_policy = retry_policy

    def ready(self):
        super().ready()

        PeriodicTask = apps.get_model(
            app_label='django_celery_beat', model_name='PeriodicTask'
        )

        try:
            self.check_broker_connectivity()
        except Exception as exception:
            print(
                'Error checking Celery broker connectivity: {}'.format(exception)
            )
            exit(1)

        try:
            self.check_results_backend_connectivity()
        except Exception as exception:
            print(
                'Error checking Celery result backend connectivity: {}'.format(exception)
            )
            exit(1)

        CeleryQueue.load_modules()

        if settings.DEBUG or settings.TESTING:
            PeriodicTask.add_to_class(
                name='save', value=factory_method_periodic_task_save(
                    super_save=PeriodicTask.save
                )
            )

        SourceColumn(
            attribute='label', is_identifier=True, label=_('Label'),
            source=CeleryQueue
        )
        SourceColumn(
            attribute='name', include_label=True, label=_('Name'),
            source=CeleryQueue
        )
        SourceColumn(
            attribute='default_queue', include_label=True,
            label=_('Default queue?'), source=CeleryQueue,
            widget=TwoStateWidget
        )
        SourceColumn(
            attribute='transient', help_text=_(
                'Transient queues are not persistent. Tasks in a transient '
                'queue are lost if the broker is restarted. Transient '
                'queues use less resources and managed non critical tasks.'
            ), include_label=True, label=_('Is transient?'),
            source=CeleryQueue, widget=TwoStateWidget
        )
        SourceColumn(
            attribute='get_task_type_count', source=CeleryQueue
        )

        SourceColumn(
            attribute='label', label=_('Label'), source=TaskType
        )
        SourceColumn(
            attribute='name', label=_('Name'), source=TaskType
        )
        SourceColumn(
            attribute='dotted_path', label=_('Dotted path'), source=TaskType
        )
        SourceColumn(
            attribute='schedule', label=_('Schedule'), source=TaskType
        )

        SourceColumn(
            attribute='task_type', include_label=True, label=_('Type'),
            source=Task
        )
        SourceColumn(
            attribute='get_time_started', include_label=True,
            label=_('Start time'), source=Task
        )
        SourceColumn(
            func=lambda context: context['object'].kwargs['hostname'],
            include_label=True, label=_('Host'), source=Task
        )
        SourceColumn(
            func=lambda context: context['object'].kwargs['args'],
            include_label=True, label=_('Arguments'), source=Task
        )
        SourceColumn(
            func=lambda context: context['object'].kwargs['kwargs'],
            include_label=True, label=_('Keyword arguments'), source=Task
        )
        SourceColumn(
            func=lambda context: context['object'].kwargs['worker_pid'],
            include_label=True, label=_('Worker process ID'), source=Task
        )

        SourceColumn(
            attribute='label', is_identifier=True, label=_('Label'),
            source=Worker
        )
        SourceColumn(
            attribute='name', label=_('Name'), source=Worker
        )
        SourceColumn(
            attribute='maximum_memory_per_child', help_text=_(
                'Maximum amount of resident memory a worker can execute '
                'before it\'s replaced by a new process.'
            ), label=_('Maximum memory per child'), source=Worker
        )
        SourceColumn(
            attribute='maximum_tasks_per_child', help_text=_(
                'Maximum number of tasks a worker can execute before it\'s '
                'replaced by a new process.'
            ), label=_('Maximum tasks per child'), source=Worker
        )
        SourceColumn(
            attribute='concurrency', help_text=_(
                'The number of worker processes/threads to launch. '
                'Defaults to the number of CPUs available on the machine.'
            ), label=_('Concurrency'), source=Worker
        )
        SourceColumn(
            attribute='nice_level', help_text=_(
                'The nice value determines the priority of the process. '
                'A higher value lowers the priority. The default '
                'value is 0.'
            ), label=_('Nice level'), source=Worker
        )
        SourceColumn(attribute='get_queue_count', source=Worker)

        menu_list_facet.bind_links(
            links=(link_queue_task_type_list,), sources=(CeleryQueue,)
        )
        menu_list_facet.bind_links(
            links=(link_worker_queue_list,), sources=(Worker,)
        )
        menu_return.bind_links(
            links=(link_worker_list,),
            sources=(Worker, 'task_manager:worker_list')
        )
        menu_tools.bind_links(
            links=(link_worker_list,)
        )

        signal_perform_upgrade.connect(
            dispatch_uid='task_manager_handler_perform_upgrade',
            receiver=handler_perform_upgrade
        )



import json
import logging

from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _

from .classes import NullBackend

logger = logging.getLogger(__name__)


class QuotaBusinessModelMixin:
    def backend_label(self):
        return self.get_backend_instance().label
    backend_label.help_text = _('Driver used for this quota entry.')
    backend_label.short_description = _('Backend')

    def backend_filters(self):
        return self.get_backend_instance().filters()
    backend_filters.short_description = _('Arguments')

    def backend_usage(self):
        return self.get_backend_instance().usage()
    backend_usage.short_description = _('Usage')

    def dumps(self, data):
        self.backend_data = json.dumps(obj=data)
        self.save(
            update_fields=('backend_data',)
        )

    def get_backend_class(self):
        """
        Retrieves the backend by importing the module and the class.
        """
        try:
            return import_string(dotted_path=self.backend_path)
        except ImportError as exception:
            logger.error(exception)

            return NullBackend

    def get_backend_instance(self):
        try:
            return self.get_backend_class()(
                **self.loads()
            )
        except Exception as exception:
            logger.error(exception, exc_info=True)

            return NullBackend()

    def loads(self):
        return json.loads(s=self.backend_data)



from django.apps import apps
from django.utils.translation import ugettext_lazy as _

from mayan.apps.source_compressed.source_backends.literals import SOURCE_UNCOMPRESS_CHOICE_ASK

from .source_backends import SourceBackendWebForm


def handler_create_default_document_source(sender, **kwargs):
    Source = apps.get_model(
        app_label='sources', model_name='Source'
    )

    queryset_source_web_forms = Source.objects.filter(
        backend_path=SourceBackendWebForm.get_class_path()
    )

    if not queryset_source_web_forms.exists():
        Source.objects.create_backend(
            backend_path=SourceBackendWebForm.get_class_path(),
            backend_data={
                'uncompress': SOURCE_UNCOMPRESS_CHOICE_ASK
            }, label=_('Default')
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.permissions import (
    permission_document_type_edit, permission_document_view
)
from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import (
    icon_document_smart_link_instance_list, icon_document_type_smart_links,
    icon_smart_link_condition_create, icon_smart_link_condition_delete,
    icon_smart_link_condition_edit, icon_smart_link_condition_list,
    icon_smart_link_create, icon_smart_link_delete,
    icon_smart_link_document_type_list, icon_smart_link_edit,
    icon_smart_link_instance_detail, icon_smart_link_list,
    icon_smart_link_setup
)
from .permissions import (
    permission_smart_link_create, permission_smart_link_delete,
    permission_smart_link_edit, permission_smart_link_view
)

# Document

link_document_smart_link_instance_list = Link(
    args='resolved_object.pk',
    icon=icon_document_smart_link_instance_list,
    permissions=(permission_document_view,), text=_('Smart links'),
    view='linking:document_smart_link_instance_list'
)

# Document type

link_document_type_smart_links = Link(
    args='resolved_object.pk', icon=icon_document_type_smart_links,
    permissions=(permission_document_type_edit,), text=_('Smart links'),
    view='linking:document_type_smart_links'
)

# Smart link

link_smart_link_create = Link(
    icon=icon_smart_link_create,
    permissions=(permission_smart_link_create,),
    text=_('Create new smart link'), view='linking:smart_link_create'
)
link_smart_link_delete = Link(
    args='object.pk', icon=icon_smart_link_delete,
    permissions=(permission_smart_link_delete,), tags='dangerous',
    text=_('Delete'), view='linking:smart_link_delete'
)
link_smart_link_document_types = Link(
    args='object.pk', icon=icon_smart_link_document_type_list,
    permissions=(permission_smart_link_edit,), text=_('Document types'),
    view='linking:smart_link_document_types'
)
link_smart_link_edit = Link(
    args='object.pk', icon=icon_smart_link_edit,
    permissions=(permission_smart_link_edit,), text=_('Edit'),
    view='linking:smart_link_edit'
)
link_smart_link_instance_view = Link(
    args=('document.pk', 'object.pk',), icon=icon_smart_link_instance_detail,
    permissions=(permission_smart_link_view,), text=_('Documents'),
    view='linking:smart_link_instance_view'
)
link_smart_link_list = Link(
    icon=icon_smart_link_list, text=_('Smart links'),
    view='linking:smart_link_list'
)
link_smart_link_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='linking', model_name='SmartLink',
        object_permission=permission_smart_link_view,
        view_permission=permission_smart_link_create,
    ), icon=icon_smart_link_setup, text=_('Smart links'),
    view='linking:smart_link_list'
)

# Smart link condition

link_smart_link_condition_create = Link(
    args='object.pk', icon=icon_smart_link_condition_create,
    permissions=(permission_smart_link_edit,), text=_('Create condition'),
    view='linking:smart_link_condition_create'
)
link_smart_link_condition_delete = Link(
    args='resolved_object.pk', icon=icon_smart_link_condition_delete,
    permissions=(permission_smart_link_edit,), tags='dangerous',
    text=_('Delete'), view='linking:smart_link_condition_delete'
)
link_smart_link_condition_edit = Link(
    args='resolved_object.pk', icon=icon_smart_link_condition_edit,
    permissions=(permission_smart_link_edit,), text=_('Edit'),
    view='linking:smart_link_condition_edit'
)
link_smart_link_condition_list = Link(
    args='object.pk', icon=icon_smart_link_condition_list,
    permissions=(permission_smart_link_edit,), text=_('Conditions'),
    view='linking:smart_link_condition_list'
)



from mayan.apps.source_compressed.source_backends.mixins import SourceBackendMixinCompressed
from mayan.apps.sources.source_backends.base import SourceBackend

from ..source_backend_actions.interactive_actions import (
    SourceBackendActionInteractiveDocumentUpload,
    SourceBackendActionInteractiveDocumentFileUpload
)
from ..source_backends.mixins import SourceBackendMixinInteractive


class SourceBackendTestInteractive(
    SourceBackendMixinInteractive, SourceBackend
):
    label = 'Test interactive source backend'


class SourceBackendTestInteractiveAction(
    SourceBackendMixinCompressed, SourceBackendMixinInteractive,
    SourceBackend
):
    action_class_list = (
        SourceBackendActionInteractiveDocumentFileUpload,
        SourceBackendActionInteractiveDocumentUpload
    )
    label = 'Test interactive source backend with actions'



from django.db import migrations, models
import django.db.models.deletion

import mayan.apps.databases.model_mixins


class Migration(migrations.Migration):
    dependencies = [
        ('events', '0008_auto_20180315_0029'),
        ('document_indexing', '0026_alter_indexinstancenode_options')
    ]

    operations = [
        migrations.CreateModel(
            name='IndexTemplateEventTrigger',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True,
                        serialize=False, verbose_name='ID'
                    )
                ),
                (
                    'index_template', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='event_triggers',
                        to='document_indexing.indextemplate',
                        verbose_name='Index template'
                    )
                ),
                (
                    'stored_event_type', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='events.storedeventtype',
                        verbose_name='Event type'
                    )
                )
            ],
            options={
                'verbose_name': 'Index template event trigger',
                'verbose_name_plural': 'Index template event triggers',
                'unique_together': {
                    ('index_template', 'stored_event_type')
                }
            },
            bases=(
                mayan.apps.databases.model_mixins.ExtraDataModelMixin,
                models.Model
            )
        )
    ]



from django.apps import apps
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import (
    menu_list_facet, menu_main, menu_multi_item, menu_secondary
)
from mayan.apps.dashboards.dashboards import dashboard_administrator
from mayan.apps.databases.classes import ModelQueryFields
from mayan.apps.events.classes import ModelEventType
from mayan.apps.navigation.classes import SourceColumn

from .dashboard_widgets import DashboardWidgetTotalCheckouts
from .events import (
    event_document_auto_checked_in, event_document_checked_in,
    event_document_checked_out, event_document_forcefully_checked_in
)
from .hooks import hook_is_new_file_allowed
from .links import (
    link_check_in_document, link_check_in_document_multiple,
    link_check_out_document, link_check_out_document_multiple,
    link_check_out_info, link_check_out_list
)
from .methods import (
    method_check_in, method_get_check_out_info, method_get_check_out_state,
    method_is_checked_out
)
from .permissions import (
    permission_document_check_in, permission_document_check_in_override,
    permission_document_check_out, permission_document_check_out_detail_view
)


class CheckoutsApp(MayanAppConfig):
    app_namespace = 'checkouts'
    app_url = 'checkouts'
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.checkouts'
    verbose_name = _('Checkouts')

    def ready(self):
        super().ready()

        CheckedOutDocument = self.get_model(model_name='CheckedOutDocument')
        DocumentCheckout = self.get_model(model_name='DocumentCheckout')
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )

        Document.add_to_class(name='check_in', value=method_check_in)
        Document.add_to_class(
            name='get_check_out_info', value=method_get_check_out_info
        )
        Document.add_to_class(
            name='get_check_out_state', value=method_get_check_out_state
        )
        Document.add_to_class(
            name='is_checked_out', value=method_is_checked_out
        )

        DocumentFile.register_pre_create_hook(
            func=hook_is_new_file_allowed
        )

        ModelEventType.register(
            model=Document, event_types=(
                event_document_auto_checked_in, event_document_checked_in,
                event_document_checked_out,
                event_document_forcefully_checked_in
            )
        )

        ModelPermission.register(
            model=Document, permissions=(
                permission_document_check_out,
                permission_document_check_in,
                permission_document_check_in_override,
                permission_document_check_out_detail_view
            )
        )
        ModelPermission.register_inheritance(
            model=DocumentCheckout, related='document'
        )

        model_query_fields_document = ModelQueryFields(model=Document)
        model_query_fields_document.add_select_related_field(
            field_name='documentcheckout'
        )

        SourceColumn(
            attribute='get_user_display', include_label=True, order=99,
            source=CheckedOutDocument
        )
        SourceColumn(
            attribute='get_checkout_datetime', include_label=True, order=99,
            source=CheckedOutDocument
        )
        SourceColumn(
            attribute='get_checkout_expiration', include_label=True,
            order=99, source=CheckedOutDocument
        )

        dashboard_administrator.add_widget(
            widget=DashboardWidgetTotalCheckouts, order=-1
        )

        menu_list_facet.bind_links(
            links=(link_check_out_info,), sources=(Document,)
        )
        menu_main.bind_links(
            links=(link_check_out_list,), position=40
        )
        menu_multi_item.bind_links(
            links=(
                link_check_in_document_multiple,
            ), sources=(CheckedOutDocument,)
        )
        menu_multi_item.bind_links(
            links=(
                link_check_in_document_multiple,
                link_check_out_document_multiple,
            ), sources=(Document,)
        )
        menu_multi_item.unbind_links(
            links=(
                link_check_out_document_multiple,
            ), sources=(CheckedOutDocument,)
        )
        menu_secondary.bind_links(
            links=(link_check_out_document, link_check_in_document),
            sources=(
                'checkouts:check_out_info', 'checkouts:check_out_document',
                'checkouts:check_in_document'
            )
        )



import logging

from django.contrib import messages
from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import (
    permission_document_tools, permission_document_view
)
from mayan.apps.documents.views.document_views import DocumentListView
from mayan.apps.views.generics import ConfirmView
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from .icons import (
    icon_document_duplicates_list, icon_duplicated_document_list,
    icon_duplicated_document_scan
)
from .models import DuplicateBackendEntry
from .tasks import task_duplicates_scan_all

logger = logging.getLogger(name=__name__)


class DocumentDuplicatesListView(ExternalObjectViewMixin, DocumentListView):
    external_object_permission = permission_document_view
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    view_icon = icon_document_duplicates_list

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'no_results_icon': icon_duplicated_document_list,
                'no_results_text': _(
                    'Only exact copies of this document will be shown in the '
                    'this list.'
                ),
                'no_results_title': _(
                    'There are no duplicates for this document'
                ),
                'object': self.external_object,
                'title': _(
                    'Duplicates for document: %s'
                ) % self.external_object
            }
        )
        return context

    def get_source_queryset(self):
        return DuplicateBackendEntry.objects.get_duplicates_of(
            document=self.external_object
        )


class DuplicatedDocumentListView(DocumentListView):
    view_icon = icon_duplicated_document_list

    def get_document_queryset(self):
        return DuplicateBackendEntry.objects.get_duplicated_documents(
            permission=permission_document_view, user=self.request.user
        )

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'no_results_icon': icon_duplicated_document_list,
                'no_results_text': _(
                    'Duplicates are documents that are composed of the exact '
                    'same file, down to the last byte. Files that have the '
                    'same text or OCR but are not identical or were saved '
                    'using a different file format will not appear as '
                    'duplicates.'
                ),
                'no_results_title': _(
                    'There are no duplicated documents'
                ),
                'title': _('Duplicated documents')
            }
        )
        return context


class ScanDuplicatedDocuments(ConfirmView):
    extra_context = {
        'title': _('Scan for duplicated documents?')
    }
    view_permission = permission_document_tools
    view_icon = icon_duplicated_document_scan

    def view_action(self):
        task_duplicates_scan_all.apply_async()
        messages.success(
            message=_('Duplicated document scan queued successfully.'),
            request=self.request
        )



import logging

from django.apps import apps
from django.core.files import File

from mayan.apps.sources.source_backend_actions.interfaces import (
    SourceBackendActionInterface, SourceBackendActionInterfaceRequestRESTAPI,
    SourceBackendActionInterfaceRequestViewForm,
    SourceBackendActionInterfaceTask
)
from mayan.apps.storage.compressed_files import Archive
from mayan.apps.storage.exceptions import NoMIMETypeMatch
from mayan.apps.storage.tasks import task_shared_upload_delete

from .arguments import argument_expand
from .literals import (
    SOURCE_UNCOMPRESS_CHOICE_ALWAYS, SOURCE_UNCOMPRESS_CHOICE_ASK
)

logger = logging.getLogger(name=__name__)


class SourceBackendActionMixinCompressedBase:
    def _background_task(self, expand, **kwargs):
        result = super()._background_task(**kwargs)

        if result:
            # Make this optional in case another mixin interrupted the MRO,
            # and called a background task.

            SharedUploadedFile = apps.get_model(
                app_label='storage', model_name='SharedUploadedFile'
            )

            original_server_upload_entry_list = result.pop(
                'server_upload_entry_list'
            )

            extracted_server_upload_entry_list = []

            if expand:
                for original_server_upload_entry in original_server_upload_entry_list:
                    original_server_upload_entry_extra_data = original_server_upload_entry.copy()
                    original_server_upload_entry_extra_data.pop(
                        'shared_uploaded_file_id'
                    )

                    original_shared_uploaded_file_id = original_server_upload_entry.get(
                        'shared_uploaded_file_id'
                    )

                    original_shared_uploaded_file = SharedUploadedFile.objects.get(
                        pk=original_shared_uploaded_file_id
                    )

                    try:
                        with original_shared_uploaded_file.open(mode='rb') as shared_uploaded_file_object:
                            compressed_file = Archive.open(file_object=shared_uploaded_file_object)
                            for compressed_file_member in compressed_file.members():
                                with compressed_file.open_member(filename=compressed_file_member) as compressed_file_member_file_object:
                                    shared_uploaded_file = SharedUploadedFile.objects.create(
                                        file=File(compressed_file_member_file_object)
                                    )
                                    server_upload_entry = original_server_upload_entry_extra_data.copy()
                                    server_upload_entry['shared_uploaded_file_id'] = shared_uploaded_file.pk
                                    extracted_server_upload_entry_list.append(
                                        server_upload_entry
                                    )
                    except NoMIMETypeMatch:
                        logger.debug(msg='Not expanding; Exception: NoMIMETypeMatch')
                        extracted_server_upload_entry_list.append(
                            original_server_upload_entry
                        )
                    except Exception:
                        # Cleanup on fatal errors.
                        for original_server_upload_entry in original_server_upload_entry_list:
                            task_shared_upload_delete.apply_async(
                                kwargs={
                                    'shared_uploaded_file_id': original_server_upload_entry[
                                        'shared_uploaded_file_id'
                                    ]
                                }
                            )

                        for extracted_server_upload_entry in extracted_server_upload_entry_list:
                            task_shared_upload_delete.apply_async(
                                kwargs={
                                    'shared_uploaded_file_id': extracted_server_upload_entry[
                                        'shared_uploaded_file_id'
                                    ]
                                }
                            )

                        raise
                    else:
                        original_shared_uploaded_file_id = original_server_upload_entry.get(
                            'shared_uploaded_file_id'
                        )

                        task_shared_upload_delete.apply_async(
                            kwargs={
                                'shared_uploaded_file_id': original_shared_uploaded_file_id
                            }
                        )
            else:
                extracted_server_upload_entry_list = original_server_upload_entry_list

            result['server_upload_entry_list'] = extracted_server_upload_entry_list

            return result

    def get_task_kwargs(self, expand, **kwargs):
        result = super().get_task_kwargs(**kwargs)

        result['action_interface_kwargs'].update(
            {'expand': expand}
        )

        return result


class SourceBackendActionMixinCompressedInteractive(
    SourceBackendActionMixinCompressedBase
):
    class Interface:
        class Model(SourceBackendActionInterface):
            class Argument:
                expand = argument_expand

            def process_interface_context(self):
                super().process_interface_context()

                source_backend = self.action.source.get_backend_instance()

                if source_backend.kwargs['uncompress'] == SOURCE_UNCOMPRESS_CHOICE_ASK:
                    expand = self.context['expand']
                else:
                    if source_backend.kwargs['uncompress'] == SOURCE_UNCOMPRESS_CHOICE_ALWAYS:
                        expand = True
                    else:
                        expand = False

                self.action_kwargs['expand'] = expand

        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            class Argument:
                expand = argument_expand

            def process_interface_context(self):
                super().process_interface_context()

                source_backend = self.action.source.get_backend_instance()

                if source_backend.kwargs['uncompress'] == SOURCE_UNCOMPRESS_CHOICE_ASK:
                    expand = self.context['expand']
                else:
                    if source_backend.kwargs['uncompress'] == SOURCE_UNCOMPRESS_CHOICE_ALWAYS:
                        expand = True
                    else:
                        expand = False

                self.action_kwargs['expand'] = expand

        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                expand = argument_expand

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['expand'] = self.context['expand']

        class View(SourceBackendActionInterfaceRequestViewForm):
            class Argument:
                expand = argument_expand

            def process_interface_context(self):
                super().process_interface_context()

                source_backend = self.action.source.get_backend_instance()

                if source_backend.kwargs['uncompress'] == SOURCE_UNCOMPRESS_CHOICE_ASK:
                    expand = self.context['forms']['source_form'].cleaned_data.get('expand')
                else:
                    if source_backend.kwargs['uncompress'] == SOURCE_UNCOMPRESS_CHOICE_ALWAYS:
                        expand = True
                    else:
                        expand = False

                self.action_kwargs['expand'] = expand


class SourceBackendActionMixinCompressedInteractiveNot(
    SourceBackendActionMixinCompressedBase
):
    class Interface:
        class Model(SourceBackendActionInterface):
            def process_interface_context(self):
                super().process_interface_context()

                source_backend = self.action.source.get_backend_instance()

                self.action_kwargs['expand'] = source_backend.kwargs.get('uncompress') == SOURCE_UNCOMPRESS_CHOICE_ALWAYS

        class Task(SourceBackendActionInterfaceTask):
            def process_interface_context(self):
                super().process_interface_context()

                source_backend = self.action.source.get_backend_instance()

                self.action_kwargs['expand'] = source_backend.kwargs.get('uncompress') == SOURCE_UNCOMPRESS_CHOICE_ALWAYS



from django.apps import apps

from mayan.apps.converter.exceptions import AppImageError
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerMethodAfter

from ..events import event_trashed_document_restored
from ..literals import IMAGE_ERROR_NO_VERSION_PAGES


class TrashedDocumentBusinessLogicMixin:
    @property
    def document(self):
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )

        return Document.objects.get(pk=self.pk)

    def get_api_image_url(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None
    ):
        first_page = self.pages.first()
        if first_page:
            return first_page.get_api_image_url(
                maximum_layer_order=None, transformation_instance_list=None,
                view_kwargs={'document_id': self.pk},
                viewname='rest_api:trasheddocument-image', user=user
            )
        else:
            raise AppImageError(error_name=IMAGE_ERROR_NO_VERSION_PAGES)

    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_trashed_document_restored,
        target='document'
    )
    def restore(self, user):
        self._event_actor = user
        self.in_trash = False
        # Skip the edit event at .save().
        self._event_ignore = True
        self.save(
            update_fields=('in_trash',)
        )



import json

from django.db import migrations

SOURCE_BACKEND_MAPPING_LIST = [
    {
        'backend_path': 'mayan.apps.sources.source_backends.SourceBackendIMAPEmail',
        'model_name': 'IMAPEmail'
    },
    {
        'backend_path': 'mayan.apps.sources.source_backends.SourceBackendPOP3Email',
        'model_name': 'POP3Email'
    },
    {
        'backend_path': 'mayan.apps.sources.source_backends.SourceBackendSANEScanner',
        'model_name': 'SaneScanner'
    },
    {
        'backend_path': 'mayan.apps.sources.source_backends.SourceBackendStagingFolder',
        'model_name': 'StagingFolderSource'
    },
    {
        'backend_path': 'mayan.apps.sources.source_backends.SourceBackendWatchFolder',
        'model_name': 'WatchFolderSource'
    },
    {
        'backend_path': 'mayan.apps.sources.source_backends.SourceBackendWebForm',
        'model_name': 'WebFormSource'
    }
]


def convert_source_model(apps, schema_editor, source_backend_mapping):
    Source = apps.get_model(app_label='sources', model_name='Source')
    Model = apps.get_model(
        app_label='sources', model_name=source_backend_mapping['model_name']
    )

    for source in Model.objects.using(alias=schema_editor.connection.alias).all():
        source.delete()
        Source.objects.create(
            backend_path=source_backend_mapping['backend_path'],
            backend_data=json.dumps(
                obj={
                    key: value for key, value in source.__dict__.items() if not key.startswith('_')
                }
            ), label=source.label, enabled=source.enabled
        )


def code_convert_sources(apps, schema_editor):
    for source_backend_mapping in SOURCE_BACKEND_MAPPING_LIST:
        convert_source_model(
            apps=apps, schema_editor=schema_editor,
            source_backend_mapping=source_backend_mapping
        )


class Migration(migrations.Migration):
    dependencies = [
        ('sources', '0026_auto_20201030_0253')
    ]

    operations = [
        migrations.RunPython(
            code=code_convert_sources,
            reverse_code=migrations.RunPython.noop
        )
    ]



class DeprecationWarning(UserWarning):
    """
    Warning when a feature or interface has been deprecated.
    """


class InterfaceWarning(UserWarning):
    """
    Warning when using obsolete internal interfaces.
    """



from pathlib import Path
from unittest import mock

from django.apps import apps
from django.core.files import File
from django.core.files.storage import FileSystemStorage
from django.db.models.signals import post_migrate, post_save

from mayan.apps.databases.literals import DJANGO_POSITIVE_INTEGER_FIELD_MAX_VALUE
from mayan.apps.documents.signals import signal_post_document_file_upload
from mayan.apps.documents.tests.literals import TEST_FILE_PDF_PATH
from mayan.apps.testing.tests.base import MayanMigratorTestCase

from .mixins.document_mixins import DocumentTestMixin


class DocumentsAppMigrationTestMixin(DocumentTestMixin):
    auto_create_test_document_type = False
    auto_create_test_document = False
    auto_delete_test_document_type = False

    def setUp(self):
        self.clear_signals_and_hooks()
        super().setUp()

    def tearDown(self):
        super().tearDown()
        self.restore_signals_and_hooks()

    def clear_signals_and_hooks(self):
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )

        self.document_file_hooks_pre_create = DocumentFile._hooks_pre_create
        DocumentFile._hooks_pre_create = []

        self.document_file_pre_open_hooks = DocumentFile._pre_open_hooks
        DocumentFile._pre_open_hooks = []

        self.document_file_pre_save_hooks = DocumentFile._pre_save_hooks
        DocumentFile._pre_save_hooks = []

        self.document_file_post_save_hooks = DocumentFile._post_save_hooks
        DocumentFile._post_save_hooks = []

        self.document_hooks_pre_create = Document._hooks_pre_create
        Document._hooks_pre_create = []

        self.signal_post_document_file_upload_receivers = signal_post_document_file_upload.receivers
        signal_post_document_file_upload.receivers = []
        signal_post_document_file_upload.sender_receivers_cache.clear()

        self.post_save_receivers = post_save.receivers
        post_save.receivers = []
        post_save.sender_receivers_cache.clear()

        self.post_migrate_receivers = post_migrate.receivers
        post_migrate.receivers = []
        post_migrate.sender_receivers_cache.clear()

    def restore_signals_and_hooks(self):
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )

        DocumentFile._hooks_pre_create = self.document_file_hooks_pre_create
        DocumentFile._pre_open_hooks = self.document_file_pre_open_hooks
        DocumentFile._pre_save_hooks = self.document_file_pre_save_hooks
        DocumentFile._post_save_hooks = self.document_file_post_save_hooks
        Document._hooks_pre_create = self.document_hooks_pre_create

        signal_post_document_file_upload.receivers = self.signal_post_document_file_upload_receivers
        signal_post_document_file_upload.sender_receivers_cache.clear()

        post_save.receivers = self.post_save_receivers
        post_save.sender_receivers_cache.clear()

        post_migrate.receivers = self.post_migrate_receivers
        post_migrate.sender_receivers_cache.clear()


class DocumentFileSizeCopyMigrationTestCase(
    DocumentsAppMigrationTestMixin, MayanMigratorTestCase
):
    migrate_from = ('documents', '0079_documentfile_size')
    migrate_to = ('documents', '0080_populate_file_size')

    def tearDown(self):
        name = self._test_document_file.file.name
        self._test_document_file.file.close()
        self._test_document_file.file.storage.delete(name=name)

        super().tearDown()

    def prepare(self):
        DocumentType = self.old_state.apps.get_model(
            app_label='documents', model_name='DocumentType'
        )
        Document = self.old_state.apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentFile = self.old_state.apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )

        def get_upload_filename(self, instance, filename):
            return 'test-document-file'

        DocumentType.get_upload_filename = get_upload_filename

        self._test_document_type = DocumentType.objects.create(
            label='test document type'
        )
        self._test_document = Document.objects.create(
            document_type=self._test_document_type, label='test document'
        )

        with open(file=TEST_FILE_PDF_PATH, mode='rb') as file_object:
            self._test_document_file = DocumentFile(
                document=self._test_document, file=File(file=file_object),
                filename=Path(file_object.name).name
            )
            self._test_document_file.save()

        name = self._test_document_file.file.name
        self._test_document_file.file.close()
        self._test_document_file_size = self._test_document_file.file.storage.size(
            name=name
        )

    def test_document_file_size_copy(self):
        DocumentFile = self.new_state.apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )

        self.assertEqual(
            DocumentFile.objects.first().size, self._test_document_file_size
        )


class DocumentFileSizeCopyMaxValueMigrationTestCase(
    DocumentsAppMigrationTestMixin, MayanMigratorTestCase
):
    migrate_from = ('documents', '0079_documentfile_size')
    migrate_to = ('documents', '0080_populate_file_size')

    def setUp(self):
        self.patcher = mock.patch.object(FileSystemStorage, 'size')
        self.mock_size = self.patcher.start()
        self.mock_size.return_value = DJANGO_POSITIVE_INTEGER_FIELD_MAX_VALUE + 1
        super().setUp()

    def tearDown(self):
        self.patcher.stop()

        name = self._test_document_file.file.name
        self._test_document_file.file.close()
        self._test_document_file.file.storage.delete(name=name)

        super().tearDown()

    def prepare(self):
        DocumentType = self.old_state.apps.get_model(
            app_label='documents', model_name='DocumentType'
        )
        Document = self.old_state.apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentFile = self.old_state.apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )

        def get_upload_filename(self, instance, filename):
            return 'test-document-file'

        DocumentType.get_upload_filename = get_upload_filename

        self._test_document_type = DocumentType.objects.create(
            label='test document type'
        )
        self._test_document = Document.objects.create(
            document_type=self._test_document_type, label='test document'
        )

        with open(file=TEST_FILE_PDF_PATH, mode='rb') as file_object:
            self._test_document_file = DocumentFile(
                document=self._test_document, file=File(file=file_object),
                filename=Path(file_object.name).name
            )
            self._test_document_file.save()

        self._test_document_file.file.close()

    def test_document_file_size_max_value(self):
        DocumentFile = self.new_state.apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )

        self.assertEqual(
            DocumentFile.objects.first().size, DJANGO_POSITIVE_INTEGER_FIELD_MAX_VALUE
        )


class DocumentVersionActiveMigrationTestCase(
    DocumentsAppMigrationTestMixin, MayanMigratorTestCase
):
    migrate_from = ('documents', '0055_auto_20200814_0626')
    migrate_to = ('documents', '0075_delete_duplicateddocumentold')

    def prepare(self):
        DocumentType = self.old_state.apps.get_model(
            app_label='documents', model_name='DocumentType'
        )
        Document = self.old_state.apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentVersion = self.old_state.apps.get_model(
            app_label='documents', model_name='DocumentVersion'
        )
        DocumentPage = self.old_state.apps.get_model(
            app_label='documents', model_name='DocumentPage'
        )

        self._test_document_type = DocumentType.objects.create(
            label='test document type'
        )
        self._test_document = Document.objects.create(
            document_type=self._test_document_type, label='test document'
        )

        self._test_document_version_list = []
        self._test_document_version_list.append(
            DocumentVersion.objects.create(document_id=self._test_document.pk)
        )
        self._test_document_version_list.append(
            DocumentVersion.objects.create(document_id=self._test_document.pk)
        )

        DocumentPage.objects.create(
            document_version_id=self._test_document_version_list[0].pk
        )
        DocumentPage.objects.create(
            document_version_id=self._test_document_version_list[1].pk
        )

    def test_single_active_version(self):
        DocumentFile = self.new_state.apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )
        DocumentVersion = self.new_state.apps.get_model(
            app_label='documents', model_name='DocumentVersion'
        )

        self.assertEqual(
            DocumentFile.objects.count(), 2
        )

        self.assertEqual(
            DocumentVersion.objects.count(), 2
        )

        self.assertEqual(
            DocumentVersion.objects.filter(active=True).count(), 1
        )



from django.apps import apps
from django.utils.translation import ugettext_lazy as _


def method_document_get_metadata(self, permission, user):
    AccessControlList = apps.get_model(
        app_label='acls', model_name='AccessControlList'
    )
    DocumentMetadata = apps.get_model(
        app_label='metadata', model_name='DocumentMetadata'
    )

    return AccessControlList.objects.restrict_queryset(
        permission=permission,
        queryset=DocumentMetadata.objects.filter(document=self), user=user
    )


method_document_get_metadata.help_text = _(
    'Return the metadata of the document.'
)
method_document_get_metadata.short_description = 'get_metadata()'
